"""
Controller implementation for the logic to manage a TemporaryPatch resource type
"""

# Standard
from typing import List, Type, Union

# First Party
import alog

# Local
from ..controller import Controller
from ..decorator import controller
from ..session import Session
from .temporary_patch_component import TemporaryPatchComponent

log = alog.use_channel("PATCH")


@controller(
    group="oper8.org",
    version="v1",
    kind="TemporaryPatch",
    extra_properties={"disable_vcs": True},
)
class TemporaryPatchController(Controller):
    """
    The TemporaryPatchController is a custom Controller implementation that
    manages temporary patch resources for `oper8`.

    See the ADR for full details:
        docs/adr/03-patches.md

    NOTE: When deploying using OLM, there is a requirement that only a single
        operator in a given namespace control each group/version/kind. In this
        case, it will be necessary for the consuming operator to customize the
        group/version/kind of the TemporaryPatch resource so that it is specific
        to the given operator's product. This can be accomplished by deriving a
        child of TemporaryPatchController as follows:

    ```py
    from oper8 import TemporaryPatchController, controller

    @controller(group="my.group.name", version="v1", kind="MyTemporaryPatch")
    class MyTemporaryPatchController(TemporaryPatchController):
        '''The temporary patch class for my operator!'''
    ```
    """

    def __init__(
        self, patchable_kinds: List[Union[str, Type[Controller]]], *args, **kwargs
    ):
        """Construct with the list of kinds that can be patched by this
        controller

        Args:
            patchable_kinds:  List[Union[str, Type[Controller]]]
                This list holds the kinds of resources that can be patched by
                this controller. Entries may be just a `kind` in which case any
                apiVersion will be accepted, or `apiVersion/kind` in which case
                the kind matching will be scoped to the given apiVersion. If a
                Controller type is given, its group/version/kind will be used.
        """
        self._patchable_kinds = [
            self._patchable_kind_label(entry) for entry in patchable_kinds
        ]
        super().__init__(*args, **kwargs)

    def setup_components(self, session: Session):
        """Set up the component that will apply the patch to the patched
        resource if the resource is one that can be patched.
        """
        self._do_setup_components(session, is_finalizer=False)

    def finalize_components(self, session: Session):
        """Set up the component that will remove the patch from the target"""
        self._do_setup_components(session, is_finalizer=True)

    ## Shared Utilities ########################################################

    ## Implementation Details ##################################################

    @staticmethod
    def _patchable_kind_label(entry: Union[str, Type[Controller]]) -> str:
        """This helper implements the logic for registering a controller as a
        patchable kind
        """
        if isinstance(entry, type) and issubclass(entry, Controller):
            return "/".join(
                [
                    entry.group,
                    entry.version,
                    entry.kind,
                ]
            )
        return entry

    def _do_setup_components(self, session: Session, is_finalizer: bool):
        """We set up the same set of components, but configure differently when
        running the finalizer
        """

        # Pull out the information required to look up the resource
        target_api_version = session.spec.apiVersion
        target_kind = session.spec.kind
        target_name = session.spec.name
        namespace = session.namespace
        patch_name = session.name
        assert (
            target_api_version is not None
        ), "Missing required field 'spec.apiVersion'"
        assert target_kind is not None, "Missing required field 'spec.kind'"
        assert target_name is not None, "Missing required field 'spec.name'"
        assert namespace is not None, "No namespace found in metadata"
        assert patch_name is not None, "No name found in metadata"
        log.info(
            "Processing patch for [%s/%s/%s/%s]",
            namespace,
            target_api_version,
            target_kind,
            target_name,
        )

        # Only add the component if this kind (or apiVersion/kind) is patchable
        if (
            target_kind in self._patchable_kinds
            or f"{target_api_version}/{target_kind}" in self._patchable_kinds
        ):
            log.debug(
                "Adding patch component for patchable kind %s/%s",
                target_api_version,
                target_kind,
            )
            TemporaryPatchComponent(
                session=session,
                disabled=is_finalizer,
                patch_name=patch_name,
                target_api_version=target_api_version,
                target_kind=target_kind,
                target_name=target_name,
            )



"""
Tests for the TemporaryPatchComponent
"""

# Standard
from datetime import datetime
import json

# Third Party
import pytest

# First Party
import aconfig

# Local
from oper8.constants import TEMPORARY_PATCHES_ANNOTATION_NAME
from oper8.exceptions import ClusterError, ConfigError, PreconditionError
from oper8.patch import STRATEGIC_MERGE_PATCH
from oper8.temporary_patch.temporary_patch_component import TemporaryPatchComponent
from oper8.test_helpers.helpers import (
    FailOnce,
    MockDeployManager,
    make_patch,
    setup_cr,
    setup_session,
)

## Happy Path (enabled) ########################################################


def test_enabled_new_patch():
    """Test that adding a new patch works"""
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {"name": "foo", "namespace": "ns"},
        }
    )
    patch_name = "some-patch"
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.deploy(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    assert patch_name in patch_anno_content
    assert "timestamp" in patch_anno_content[patch_name]
    assert patch_anno_content[patch_name]["api_version"] == patch.apiVersion
    assert patch_anno_content[patch_name]["kind"] == patch.kind


def test_enabled_already_present():
    """Test that when a patch is already present, the patch is not changed"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    existing_anno = {
        patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        }
    }
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(existing_anno),
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.deploy(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    assert patch_anno_content == existing_anno


def test_enabled_multi_patch():
    """Test that when an existing patch is present and a new patch is added,
    they both persist
    """
    old_patch_name = "test1"
    new_patch_name = "test2"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    existing_anno = {
        old_patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        }
    }
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(existing_anno),
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=new_patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=new_patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.deploy(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    assert len(patch_anno_content) == 2
    assert old_patch_name in patch_anno_content
    assert new_patch_name in patch_anno_content


## Happy Path (disabled) #######################################################


def test_disabled_patch_exists():
    """Test that removing a patch which exists works"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    existing_anno = {
        patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        }
    }
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(existing_anno),
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=True,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.disable(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    assert patch_anno_content == {}


def test_disabled_multi_patch():
    """Test that removing a patch from a resource with multiple patches only
    removes the desired patch
    """
    old_patch_name = "test1"
    new_patch_name = "test2"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    existing_anno = {
        old_patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        },
        new_patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        },
    }
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(existing_anno),
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=new_patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=True,
        patch_name=new_patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.disable(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    existing_anno.pop(new_patch_name)
    assert patch_anno_content == existing_anno


def test_disabled_resource_not_found():
    """Test that if the resource is removed before the patch, it's a no-op"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    namespace = "ns"
    dm = MockDeployManager()
    session = setup_session(
        deploy_manager=dm,
        full_cr=setup_cr(
            namespace=namespace,
            api_version=patch_api_version,
            kind=patch_kind,
            name=patch_name,
        ),
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=True,
        patch_name=patch_name,
        target_api_version="foo.bar/v1",
        target_kind="Foo",
        target_name=namespace,
    )
    comp.disable(session)


def test_disabled_resource_annotation_not_there():
    """Test that if the annotation has already been removed, it's a no-op"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=True,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.disable(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    assert updated_target.metadata.annotations is None


def test_disabled_resource_patch_not_there():
    """Test that if the patch has already been removed, it's a no-op"""
    old_patch_name = "test1"
    new_patch_name = "test2"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    existing_anno = {
        old_patch_name: {
            "api_version": patch_api_version,
            "kind": patch_kind,
            "timestamp": datetime.now().isoformat(),
        },
    }
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: json.dumps(existing_anno),
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=new_patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=True,
        patch_name=new_patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    comp.disable(session)
    updated_target = dm.get_obj(
        kind=target.kind,
        api_version=target.apiVersion,
        namespace=target.metadata.namespace,
        name=target.metadata.name,
    )
    assert updated_target is not None
    patch_anno = updated_target.metadata.annotations[TEMPORARY_PATCHES_ANNOTATION_NAME]
    assert patch_anno is not None
    patch_anno_content = json.loads(patch_anno)
    assert patch_anno_content == existing_anno


## Error Cases #################################################################


def test_enabled_resource_not_found():
    """Test that if the target resource is not found, the patch is rejected"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    namespace = "ns"
    dm = MockDeployManager()
    session = setup_session(
        deploy_manager=dm,
        full_cr=setup_cr(
            namespace=namespace,
            api_version=patch_api_version,
            kind=patch_kind,
            name=patch_name,
        ),
    )
    with pytest.raises(PreconditionError):
        TemporaryPatchComponent(
            session=session,
            disabled=False,
            patch_name=patch_name,
            target_api_version="foo.bar/v1",
            target_kind="Foo",
            target_name=namespace,
        )


def test_enabled_resource_lookup_error():
    """Test that if the resource errors on lookup, an error is raised"""
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    namespace = "ns"
    dm = MockDeployManager(get_state_fail=FailOnce((False, None), fail_number=2))
    session = setup_session(
        deploy_manager=dm,
        full_cr=setup_cr(
            namespace=namespace,
            api_version=patch_api_version,
            kind=patch_kind,
            name=patch_name,
        ),
    )
    with pytest.raises(ClusterError):
        TemporaryPatchComponent(
            session=session,
            disabled=False,
            patch_name=patch_name,
            target_api_version="foo.bar/v1",
            target_kind="Foo",
            target_name=namespace,
        )


def test_enabled_annotation_not_json():
    """Test that if the target resource has a non-json value for the annotation,
    an error is raised
    """
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: "{",
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    with pytest.raises(ConfigError):
        comp.deploy(session)


def test_enabled_annotation_not_dict():
    """Test that if the target resource has a value that is json, but not a dict
    for the annotation, an error is raised
    """
    patch_name = "some-patch"
    patch_api_version = "foo.bar.com/v2"
    patch_kind = "TestTemporaryPatch"
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {
                "name": "foo",
                "namespace": "ns",
                "annotations": {
                    TEMPORARY_PATCHES_ANNOTATION_NAME: "some string",
                },
            },
        }
    )
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
        api_version=patch_api_version,
        kind=patch_kind,
    )
    dm = MockDeployManager(resources=[target])
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    with pytest.raises(ConfigError):
        comp.deploy(session)


def test_enabled_deploy_error():
    """Test that if redeploying the resource fails, an error is raised"""
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {"name": "foo", "namespace": "ns"},
        }
    )
    patch_name = "some-patch"
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
    )
    dm = MockDeployManager(resources=[target], deploy_fail=True)
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    with pytest.raises(ClusterError):
        comp.deploy(session)


def test_enabled_deploy_raise():
    """Test that if redeploying the resource raises, an error is raised"""
    target = aconfig.Config(
        {
            "kind": "Foo",
            "apiVersion": "v1",
            "metadata": {"name": "foo", "namespace": "ns"},
        }
    )
    patch_name = "some-patch"
    patch = make_patch(
        STRATEGIC_MERGE_PATCH,
        {"foo": {"key": "value"}},
        namespace=target.metadata.namespace,
        target=target,
        name=patch_name,
    )
    dm = MockDeployManager(resources=[target], deploy_raise=True)
    session = setup_session(
        deploy_manager=dm,
        namespace=target.metadata.namespace,
        full_cr=patch,
    )
    comp = TemporaryPatchComponent(
        session=session,
        disabled=False,
        patch_name=patch_name,
        target_api_version=target.apiVersion,
        target_kind=target.kind,
        target_name=target.metadata.name,
    )
    with pytest.raises(AssertionError):
        comp.deploy(session)



"""
Test the top-level ansible module
"""

# Standard
from contextlib import contextmanager
from unittest import mock
import functools
import os
import tempfile

# Third Party
from openshift.dynamic import DynamicClient, EagerDiscoverer
import pytest

# First Party
import alog

# Local
from oper8 import Controller, constants
from oper8.log_format import Oper8JsonFormatter
from oper8.test_helpers.helpers import (
    DummyController,
    ModuleExit,
    TestRecorder,
    maybe_temp_dir,
    mock_config_file,
    mock_sys_exit,
)
from oper8.test_helpers.kub_mock import mock_kub_client_constructor
from oper8.utils import merge_configs

## Helpers #####################################################################

log = alog.use_channel("TEST")


DEFAULT_CONTROLLER_IMPORT = "test_k8s_application.TestController"


class AppTestRecorder(TestRecorder):
    def __init__(self, mock_client=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.mock_client = (
            DynamicClient(mock_client, discoverer=EagerDiscoverer)
            if mock_client
            else None
        )


@contextmanager
def mock_app_calls(recorder, module_params, patch_controller_class):
    if not patch_controller_class:
        yield
    else:
        import_name = module_params.get("controller_class", DEFAULT_CONTROLLER_IMPORT)

        # Disable reimporting the controller so mocks work
        with mock.patch(
            "oper8.reconcile.ReconcileManager.__init__.__defaults__",
            (None, None, None, False),
        ):

            with mock.patch(
                ".".join([import_name, "log_constructed"]),
                lambda _, config_defaults: recorder.add_event(
                    recorder.CONSTRUCTED, config_defaults
                ),
            ):
                with mock.patch(
                    ".".join([import_name, "setup_components"]),
                    lambda _, deploy_config: recorder.add_event(
                        recorder.SETUP_COMPONENTS, deploy_config
                    ),
                ):
                    with mock.patch(
                        ".".join([import_name, "finalize_components"]),
                        lambda _, deploy_config: recorder.add_event(
                            recorder.FINALIZE_COMPONENTS, deploy_config
                        ),
                    ):
                        log.debug("Yielding from patched app calls")
                        yield


DEFAULT_VERSION = "v1.2.3.test"


def make_cr(overrides=None):
    overrides = overrides or {}
    return merge_configs(
        {
            "apiVersion": "test.foo.bar.com/v1",
            "kind": "Foo",
            "metadata": {
                "name": "foo",
                "namespace": "bar",
                "uid": "12345678",
            },
            "spec": {"version": DEFAULT_VERSION},
        },
        overrides,
    )


class TestController(Controller):
    """We aren't testing the controller stack, so this is a minimal stub"""

    group = "test.com"
    version = "v2"
    kind = "Test"

    # This is not a test class!
    __test__ = False

    def __init__(self, config_defaults=None, **kwargs):
        super().__init__(config_defaults=config_defaults, **kwargs)
        self.log_constructed(config_defaults)

    def log_constructed(self, config_defaults):
        pass

    def setup_components(self, session):
        pass

    def finalize_components(self, session):
        pass


class BrokenTestController(TestController):
    """Test controller that will fail during rollout"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reconcile(self, *args, **kwargs):
        raise RuntimeError("Oh no!")

    def log_constructed(self, config_defaults):
        pass

    def setup_components(self, session):
        pass

    def finalize_components(self, session):
        pass


class TestControllerWithResources(DummyController):
    """Dummy controller that deploys some api resources"""

    # This is not a test class!
    __test__ = False

    def log_constructed(self, config_defaults):
        pass

    def setup_components(self, session):
        pass

    def finalize_components(self, session):
        pass

    def __init__(self, **kwargs):
        super().__init__(
            components=[
                {
                    "name": "foo-comp",
                    "api_objects": [
                        ("foo", {"kind": "Foo", "metadata": {"namespace": "bar"}})
                    ],
                }
            ],
            **kwargs,
        )

    def setup_components(self, session):
        log.debug("TestControllerWithResources.setup_components")
        super().setup_components(session)


def mock_alog_configure(rec, *args, **kwargs):
    rec.add_event("alog_configure", (args, kwargs))


def run_module(
    module_params=None,
    add_module_param_defaults=True,
    cr_overrides=None,
    config_defaults=None,
    recorder=None,
    cluster_state=None,
    mock_logging=True,
    patch_controller_class=True,
):
    # Local
    from oper8.watch_manager.ansible_watch_manager.modules.k8s_application import main

    with maybe_temp_dir() as temp_dir:
        log.info("Temp Dir: %s", temp_dir)

        cr = make_cr(cr_overrides)

        # Add the CR to the default cluster state
        cluster_state = cluster_state or {}
        cluster_state.setdefault(cr["metadata"].get("namespace"), {}).setdefault(
            cr["kind"], {}
        ).setdefault(cr["apiVersion"], {}).setdefault(cr["metadata"].get("name"), cr)

        # If this is a dry run, go with it, otherwise mock the k8s api
        with mock_kub_client_constructor(cluster_state=cluster_state) as mock_client:

            # Mock sys.exit and keep track of the exit code
            log.info("Mocking sys.exit")
            recorder = recorder or AppTestRecorder(mock_client)
            with mock_sys_exit(recorder):

                # Add the defaults to send output to the temp dir
                module_params = module_params or {}
                log.info("Module Params: %s", module_params)
                if add_module_param_defaults:
                    module_params["working_dir"] = os.path.join(temp_dir, "working")
                    module_params.setdefault(
                        "controller_class",
                        DEFAULT_CONTROLLER_IMPORT,
                    )
                    module_params.setdefault("version", DEFAULT_VERSION)
                    module_params.setdefault("full_cr", str(cr))

                # Wire the recorder into the test app
                log.info("Mocking app calls")
                with mock_app_calls(recorder, module_params, patch_controller_class):

                    # Create the config file for the module
                    config_object = {"ANSIBLE_MODULE_ARGS": module_params}

                    # Inject the config file as sys.argv[1]
                    log.info("Mocking the config file input")
                    with mock_config_file(config_object):

                        # run it!
                        try:
                            log.info("------------------")
                            if mock_logging:
                                with mock.patch(
                                    "alog.configure",
                                    functools.partial(mock_alog_configure, recorder),
                                ):
                                    main()
                            else:
                                main()
                        except ModuleExit:
                            pass

                        # Merge events from controller
                        return recorder


## Happy Path ##################################################################


@pytest.mark.ansible
def test_run_module_reconcile():
    """Test that a straightforward execution of the module works as expected"""
    recorder = run_module()
    assert recorder.retcode == 0
    assert recorder.events == [
        "alog_configure",
        AppTestRecorder.CONSTRUCTED,
        AppTestRecorder.SETUP_COMPONENTS,
        AppTestRecorder.EXITED,
    ]


@pytest.mark.ansible
def test_run_module_finalize():
    """Test that a straightforward execution of the module as a finalizer works
    as expected
    """
    recorder = run_module(module_params={"operation": "remove"})
    assert recorder.retcode == 0
    assert recorder.events == [
        "alog_configure",
        AppTestRecorder.CONSTRUCTED,
        AppTestRecorder.FINALIZE_COMPONENTS,
        AppTestRecorder.EXITED,
    ]


@pytest.mark.ansible
def test_paused():
    """Test if the paused annotation is present, the controller is not executed"""
    recorder = run_module(
        cr_overrides={
            "metadata": {"annotations": {constants.PAUSE_ANNOTATION_NAME: "true"}}
        }
    )
    assert recorder.retcode == 0
    assert recorder.events == ["alog_configure", AppTestRecorder.EXITED]


@pytest.mark.ansible
def test_log_config():
    recorder = AppTestRecorder()
    annotations = {
        constants.LOG_DEFAULT_LEVEL_NAME: "debug1",
        constants.LOG_FILTERS_NAME: "FOO:debug4",
        constants.LOG_THREAD_ID_NAME: "true",
        constants.LOG_JSON_NAME: "true",
    }

    with mock.patch("alog.configure", functools.partial(mock_alog_configure, recorder)):
        recorder = run_module(
            recorder=recorder, cr_overrides={"metadata": {"annotations": annotations}}
        )
        assert recorder.retcode == 0
        assert "alog_configure" in recorder.events
        config_args, config_kwargs = recorder.event_details["alog_configure"]
        assert not config_args
        assert (
            config_kwargs["default_level"]
            == annotations[constants.LOG_DEFAULT_LEVEL_NAME]
        )
        assert config_kwargs["filters"] == annotations[constants.LOG_FILTERS_NAME]
        assert config_kwargs["thread_id"] == True
        assert isinstance(config_kwargs["formatter"], Oper8JsonFormatter)


@pytest.mark.ansible
def test_log_file():
    """Make sure that setting a log_file results in a callable handler_generator"""
    recorder = AppTestRecorder()
    with mock.patch("alog.configure", functools.partial(mock_alog_configure, recorder)):
        with tempfile.TemporaryDirectory() as temp_log_dir:
            recorder = run_module(
                recorder=recorder,
                module_params={"log_file": os.path.join(temp_log_dir, "foo.bar.log")},
            )
            assert recorder.retcode == 0
            assert "alog_configure" in recorder.events
            config_args, config_kwargs = recorder.event_details["alog_configure"]
            assert not config_args
            assert config_kwargs["handler_generator"] is not None
            assert callable(config_kwargs["handler_generator"])


## Error Cases #################################################################


@pytest.mark.ansible
def test_required_args():
    """Test missing required arguments are caught and cause a non-zero exit"""
    recorder = run_module(
        add_module_param_defaults=False, module_params={"full_cr": "{}"}
    )
    assert recorder.retcode != 0


@pytest.mark.ansible
def test_unparsable_full_cr():
    """Test an un-parsable value for a full_cr exits with an error"""
    recorder = run_module(module_params={"full_cr": "not a dict"})
    assert recorder.retcode != 0


@pytest.mark.ansible
def test_incomplete_full_cr():
    """Test a CR which doesn't have the kubernetes basics"""
    recorder = run_module(module_params={"full_cr": "{}"})
    assert recorder.retcode == 0


@pytest.mark.ansible
def test_bad_controller_module():
    """Test that a bad module in controller_class causes an error exit"""
    recorder = run_module(
        module_params={"controller_class": "foo.bar.Baz"},
        patch_controller_class=False,
    )
    assert recorder.retcode == 0


@pytest.mark.ansible
def test_missing_controller_class():
    """Test that a class name that is missing from the module causes an error
    exit
    """
    recorder = run_module(
        module_params={"controller_class": "test_k8s_controller.NotThere"},
        patch_controller_class=False,
    )
    assert recorder.retcode == 0


@pytest.mark.ansible
def test_bad_controller_class():
    """Test that a class name that is not a class causes an error exit"""
    recorder = run_module(
        module_params={
            "controller_class": "test_k8s_controller.test_bad_controller_class",
        },
        patch_controller_class=False,
    )
    assert recorder.retcode == 0


@pytest.mark.ansible
def test_controller_exception():
    """Test that an exception in the Controller causes an error exit"""
    recorder = run_module(
        module_params={"controller_class": "test_k8s_application.BrokenTestController"}
    )
    assert recorder.retcode == 0


@pytest.mark.ansible
def test_owner_references():
    """Test that running the module creates resources with owner references"""
    recorder = run_module(
        cluster_state={"bar": {"Foo": {"v1": {}}}},
        module_params={
            "controller_class": "test_k8s_application.TestControllerWithResources"
        },
        # We don't patch the controller class here because we want the kube mock
        # to be able to fetch the deployed resources and they don't get deployed
        # if the patches take effect
        patch_controller_class=False,
    )
    assert recorder.retcode == 0

    # Fetch the content of the deployed component
    foos = recorder.mock_client.resources.get(kind="Foo", api_version="v1")
    resource = foos.get(name="foo", namespace="bar").to_dict()
    assert resource
    log.debug(resource)
    assert len(resource["metadata"]["ownerReferences"]) == 1



"""
Tests for the HeartbeatThread
"""
# Standard
from datetime import datetime, timedelta
from unittest import mock

# Local
from oper8.test_helpers.pwm_helpers import (
    MockedHeartbeatThread,
    heartbeat_file,
    read_heartbeat_file,
)

## Helpers #####################################################################


class FailOnceOpen:
    def __init__(self, fail_on: int = 1):
        self.call_num = 0
        self.fail_on = fail_on
        self._real_open = open

    def __call__(self, *args, **kwargs):
        self.call_num += 1
        if self.call_num == self.fail_on:
            print(f"Raising on call {self.call_num}")
            raise OSError("Yikes")
        print(f"Returning from call {self.call_num}")
        return self._real_open(*args, **kwargs)


## Tests #######################################################################


def test_heartbeat_happy_path(heartbeat_file):
    """Make sure the heartbeat initializes correctly"""
    hb = MockedHeartbeatThread(heartbeat_file, "1s")

    # Heartbeat not run until started
    with open(heartbeat_file) as handle:
        assert not handle.read()

    # Start and stop the thread to trigger the first heartbeat only
    hb.start_thread()
    hb.wait_for_beat()
    hb.stop_thread()

    # Make sure the heartbeat is "current"
    assert read_heartbeat_file(heartbeat_file) > (datetime.now() - timedelta(seconds=5))


def test_heartbeat_ongoing(heartbeat_file):
    """Make sure that the heartbeat continues to beat in an ongoing way"""
    hb = MockedHeartbeatThread(heartbeat_file, "1s")

    # Start the thread and read the first one
    hb.start_thread()
    hb.wait_for_beat()
    first_hb = read_heartbeat_file(heartbeat_file)

    # Wait a bit and read again
    hb.wait_for_beat()
    hb.stop_thread()
    later_hb = read_heartbeat_file(heartbeat_file)
    assert later_hb > first_hb


def test_heartbeat_with_exception(heartbeat_file):
    """Make sure that a sporadic failure does not terminate the heartbeat"""
    # Mock so that the third call to open will raise. This correlates with the
    # second heartbeat since we read the file using open after each heartbeat
    with mock.patch("builtins.open", new=FailOnceOpen(3)):
        hb = MockedHeartbeatThread(heartbeat_file, "1s")
        hb.start_thread()

        # The first beat succeeds
        hb.wait_for_beat()
        first_hb = read_heartbeat_file(heartbeat_file)

        # The first beat raises, but doesn't cause any problems
        hb.wait_for_beat()
        second_hb = read_heartbeat_file(heartbeat_file)

        # The third beat succeeds
        hb.wait_for_beat()
        third_hb = read_heartbeat_file(heartbeat_file)
        hb.stop_thread()

        assert first_hb == second_hb
        assert third_hb > first_hb



"""
Tests for the __main__.py entrypoint to the library as an executable
"""

# Standard
from datetime import datetime, timedelta
from pathlib import Path
from unittest import mock
import os
import sys
import tempfile

# Third Party
import pytest
import yaml

# First Party
import alog

# Local
from oper8 import Component, Controller, component, config, controller, watch_manager
from oper8.__main__ import main
from oper8.test_helpers.helpers import (
    ModuleExit,
    TestRecorder,
    library_config,
    mock_sys_exit,
)
from oper8.watch_manager.python_watch_manager.threads.heartbeat import HeartbeatThread

log = alog.use_channel("TEST")


## Sample Application ##########################################################
#
# NOTE: We don't use DummyController so that main doesn't try to register a
#   watch for it automatically
##


@component("foo")
class FooComponent(Component):

    CONSTRUCTED = False
    FOUND_EXTERNAL_RESOURCE = False

    def __init__(self, session):
        super().__init__(session=session)
        self.add_resource(
            self.name, {"kind": "Foo", "apiVersion": "v1", "metadata": {"name": "foo"}}
        )
        self.__class__.CONSTRUCTED = True

    def build_chart(self, session):
        if session.spec.do_check:
            log.debug("Doing the external resource check")
            success, content = session.get_object_current_state(
                kind="External",
                name="external-thing",
            )
            assert success
            if content is not None:
                self.__class__.FOUND_EXTERNAL_RESOURCE = True

    @classmethod
    def reset(cls):
        cls.CONSTRUCTED = False
        cls.FOUND_EXTERNAL_RESOURCE = False


@component("bar")
class BarComponent(Component):

    CONSTRUCTED = False

    def __init__(self, session):
        super().__init__(session=session)
        self.add_resource(
            self.name, {"kind": "Bar", "apiVersion": "v1", "metadata": {"name": "bar"}}
        )
        self.__class__.CONSTRUCTED = True

    @classmethod
    def reset(cls):
        cls.CONSTRUCTED = False


@component("subsystem")
class SubsystemComponent(Component):

    CONSTRUCTED = False

    def __init__(self, session):
        super().__init__(session=session)
        self.add_resource(
            self.name,
            {
                "kind": "Subsystem",
                "apiVersion": "foo.bar/v1",
                "metadata": {"name": "subsystem", "namespace": "default"},
            },
        )
        self.__class__.CONSTRUCTED = True

    @classmethod
    def reset(cls):
        cls.CONSTRUCTED = False


@controller(group="foo.bar", version="v1", kind="Subsystem")
class SubsystemController(Controller):

    CONSTRUCTED = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug("Setting up %s", self)
        self.__class__.CONSTRUCTED = True

    def setup_components(self, session):
        FooComponent(session)

    @classmethod
    def reset(cls):
        cls.CONSTRUCTED = False


@controller(group="foo.bar", version="v1", kind="Parent")
class ParentController(Controller):

    CONSTRUCTED = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug("Setting up %s", self)
        self.__class__.CONSTRUCTED = True

    def setup_components(self, session):
        BarComponent(session)
        SubsystemComponent(session)

    @classmethod
    def reset(cls):
        cls.CONSTRUCTED = False


## Helpers #####################################################################


@pytest.fixture(autouse=True)
def reset_sys_argv():
    """All tests need to muck with sys.argv, so this fixture will reset it to an
    empty list before each runs
    """
    with mock.patch.object(sys, "argv", ["oper8"]):
        yield


@pytest.fixture(autouse=True)
def reset():
    """Between tests we want to reset the app classes"""
    SubsystemController.reset()
    ParentController.reset()
    FooComponent.reset()
    BarComponent.reset()
    SubsystemComponent.reset()
    watch_manager.WatchManagerBase._ALL_WATCHES = {}


@pytest.fixture
def recorder():
    """All error tests need to capture sys.exit with a TestRecorder"""
    recorder = TestRecorder()
    with mock_sys_exit(recorder):
        yield recorder


## Happy Path Tests ############################################################


def test_controller_types_found():
    """Make sure that all controller types are found within the module"""
    sys.argv.extend(["--module_name", __name__])
    with library_config(dry_run=True):
        main()
    assert ParentController.CONSTRUCTED
    assert SubsystemController.CONSTRUCTED
    assert not FooComponent.CONSTRUCTED
    assert not BarComponent.CONSTRUCTED
    assert not SubsystemComponent.CONSTRUCTED


def test_single_controller_type_found():
    """Make sure that only the provided controller type is found within the module"""
    sys.argv.extend(["--module_name", __name__])
    with library_config(dry_run=True, controller_name="SubsystemController"):
        main()
    assert not ParentController.CONSTRUCTED
    assert SubsystemController.CONSTRUCTED
    assert not FooComponent.CONSTRUCTED
    assert not BarComponent.CONSTRUCTED
    assert not SubsystemComponent.CONSTRUCTED


def test_dry_run_reconcile_recursion():
    """Make sure that a reconciliation can run with a --cr is given and that it
    will recurse to the subsystem
    """
    with tempfile.NamedTemporaryFile() as cr_file:
        cr_file.write(
            yaml.safe_dump(
                {
                    "apiVersion": "foo.bar/v1",
                    "kind": "Parent",
                    "metadata": {
                        "name": "test",
                        "namespace": "test",
                    },
                }
            ).encode("utf-8")
        )
        cr_file.flush()
        sys.argv.extend(["--module_name", __name__, "--cr", cr_file.name])
        with library_config(dry_run=True):
            main()
        assert ParentController.CONSTRUCTED
        assert SubsystemController.CONSTRUCTED
        assert FooComponent.CONSTRUCTED
        assert BarComponent.CONSTRUCTED
        assert SubsystemComponent.CONSTRUCTED


def test_resource_dir():
    """Make sure that a --resource_dir can be specified and that valid yaml
    files are parsed form the directory
    """
    with tempfile.TemporaryDirectory() as resource_dir:

        # Put a valid yaml file for the "external" resource in the resource dir
        with open(os.path.join(resource_dir, "external.yaml"), "w") as handle:
            handle.write(
                yaml.safe_dump(
                    {
                        "apiVersion": "something.external",
                        "kind": "External",
                        "metadata": {
                            "name": "external-thing",
                            "namespace": "test",
                        },
                    }
                )
            )

        # Put another non-yaml file in there too to make sure that only valid
        # yaml is read
        with open(os.path.join(resource_dir, "README.md"), "w") as handle:
            handle.write("# External Stuff\nThis has some external stuff!")

        with tempfile.NamedTemporaryFile("w") as cr_file:
            cr_file.write(
                yaml.safe_dump(
                    {
                        "apiVersion": "foo.bar/v1",
                        "kind": "Subsystem",
                        "metadata": {
                            "name": "test",
                            "namespace": "test",
                        },
                        "spec": {"do_check": True},
                    }
                )
            )
            cr_file.flush()
            sys.argv.extend(
                [
                    "--module_name",
                    __name__,
                    "--cr",
                    cr_file.name,
                    "--resource_dir",
                    resource_dir,
                ]
            )
            with library_config(dry_run=True):
                main()
            assert SubsystemController.CONSTRUCTED
            assert FooComponent.CONSTRUCTED
            assert FooComponent.FOUND_EXTERNAL_RESOURCE


def test_supported_versions():
    """Make sure that the list of supported versions can be set via a command
    line argument
    """
    sys.argv.extend(
        [
            "--module_name",
            __name__,
            "--supported_versions",
            "foo",
            "bar",
        ]
    )
    with library_config(dry_run=True):
        main()
        assert config.supported_versions == ["foo", "bar"]


def test_dry_run_arg():
    """Make sure that the --dry_run arg works as expected"""
    sys.argv.extend(
        [
            "--module_name",
            __name__,
            "--dry_run",
        ]
    )
    main()


## Error Case Tests ############################################################


def test_module_name_required(recorder):
    """Make sure the --module_name argument is required"""
    with pytest.raises(ModuleExit):
        main()
    assert recorder.EXITED in recorder.events
    assert recorder.retcode not in [None, 0]


def test_resource_dir_not_found(recorder):
    """Make sure the --resource_dir argument is required to point to a valid
    directory
    """
    sys.argv.extend(["--module_name", __name__, "--resource_dir", "some/bad/path"])
    with library_config(dry_run=True):
        with pytest.raises(AssertionError):
            main()


def test_resource_dir_to_file(recorder):
    """Make sure the --resource_dir argument cannot point to a file"""
    with tempfile.NamedTemporaryFile("w") as handle:
        handle.write("some stuff")
        handle.flush()
        sys.argv.extend(["--module_name", __name__, "--resource_dir", handle.name])
        with library_config(dry_run=True):
            with pytest.raises(AssertionError):
                main()


def test_cr_not_found(recorder):
    """Make sure the --cr argument must point to a file that exists"""
    sys.argv.extend(["--module_name", __name__, "--cr", "some/bad/file.yaml"])
    with library_config(dry_run=True):
        with pytest.raises(AssertionError):
            main()


def test_cr_not_yaml(recorder):
    """Make sure the --cr argument must point to a file that is valid yaml"""
    with tempfile.NamedTemporaryFile("w") as handle:
        handle.write("{not\nyaml\n  really")
        handle.flush()
        sys.argv.extend(["--module_name", __name__, "--cr", handle.name])
        with library_config(dry_run=True):
            with pytest.raises(yaml.parser.ParserError):
                main()


def test_cr_to_dir(recorder):
    """Make sure the --cr argument cannot point to a directory"""
    with tempfile.TemporaryDirectory() as some_dir:
        sys.argv.extend(["--module_name", __name__, "--cr", some_dir])
        with library_config(dry_run=True):
            with pytest.raises(AssertionError):
                main()


def test_cr_without_dry_run(recorder):
    """Make sure the --cr argument can only be specified in dry_run"""
    with tempfile.NamedTemporaryFile("w") as handle:
        handle.write(yaml.safe_dump({"foo": "bar"}))
        handle.flush()
        sys.argv.extend(["--module_name", __name__, "--cr", handle.name])
        with library_config(dry_run=False):
            with pytest.raises(AssertionError):
                main()


def test_resource_dir_without_dry_run(recorder):
    """Make sure the --resource_dir argument can only be specified in dry_run"""
    with tempfile.TemporaryDirectory() as some_dir:
        sys.argv.extend(["--module_name", __name__, "--resource_dir", some_dir])
        with library_config(dry_run=False):
            with pytest.raises(AssertionError):
                main()


def test_invalid_single_controller_type_not_found():
    """Ensure that an AttributeError is thrown when given an invalid controller"""
    sys.argv.extend(["--module_name", __name__])

    with library_config(dry_run=False, controller_name="FooBar"):
        with pytest.raises(AttributeError):
            main()
    assert not ParentController.CONSTRUCTED
    assert not SubsystemController.CONSTRUCTED
    assert not FooComponent.CONSTRUCTED
    assert not BarComponent.CONSTRUCTED
    assert not SubsystemComponent.CONSTRUCTED


############################### Health Check CMD ###############################
def test_valid_health_check():
    """Ensure that the health check"""
    with tempfile.NamedTemporaryFile() as named_temp_file:
        file_path = Path(named_temp_file.name)
        file_path.write_text(datetime.now().strftime(HeartbeatThread._DATE_FORMAT))

        sys.argv.extend(["check-heartbeat", "--file", str(file_path), "--delta", "120"])
        main()


def test_to_old_health_check():
    """Ensure that the health check"""
    with tempfile.NamedTemporaryFile() as named_temp_file:
        file_path = Path(named_temp_file.name)

        old_time = datetime.now() - timedelta(seconds=100)
        file_path.write_text(old_time.strftime(HeartbeatThread._DATE_FORMAT))

        sys.argv.extend(["check-heartbeat", "--file", str(file_path), "--delta", "10"])
        with pytest.raises(KeyError):
            main()


def test_no_file_health_check():
    """Ensure that the health check"""
    sys.argv.extend(["check-heartbeat", "--file", "/some/file", "--delta", "10"])
    with pytest.raises(FileNotFoundError):
        main()



"""
This module holds helpers that rely on oper8.x
"""
# Standard
import os

# Local
from .helpers import TEST_DATA_DIR
from oper8.session import Session
from oper8.x.utils import common
from oper8.x.utils.tls_context.internal import InternalCaComponent


def set_object_test_state(
    session: Session,
    kind: str,
    name: str,
    value,
    scoped_name: bool = True,
    parent_component_name: str = "test",
):
    obj_name = (
        common.get_resource_cluster_name(
            resource_name=name,
            component=parent_component_name,
            session=session,
        )
        if scoped_name
        else name
    )
    value.setdefault("metadata", {}).setdefault("name", obj_name)
    value.setdefault("metadata", {}).setdefault(
        "namespace", session.namespace or "default"
    )
    value.setdefault("kind", kind)
    success, changed = session.deploy_manager.deploy([value])
    assert success, "Failed to set test state!"
    return success, changed


def set_secret_data(
    session: Session,
    name,
    data=None,
    string_data=None,
    secret_type="generic",
    scoped_name: bool = True,
):
    set_object_test_state(
        session,
        "Secret",
        name,
        {"type": secret_type, "data": data, "stringData": string_data},
        scoped_name=scoped_name,
    )


def set_tls_ca_secret(session):
    """Set the key/cert content for the shared CA secret. This function returns
    the pem-encoded values for convenience in other tests
    """
    with open(os.path.join(TEST_DATA_DIR, "test_ca.key")) as f:
        key_pem = f.read()
    with open(os.path.join(TEST_DATA_DIR, "test_ca.crt")) as f:
        crt_pem = f.read()
    set_secret_data(
        session,
        InternalCaComponent.CA_SECRET_NAME,
        data={
            InternalCaComponent.CA_KEY_FILENAME: common.b64_secret(key_pem),
            InternalCaComponent.CA_CRT_FILENAME: common.b64_secret(crt_pem),
        },
    )

    return key_pem, crt_pem



"""
Tests for the Python Watch Manager
"""
# Standard
from contextlib import contextmanager
import copy
import time

# Third Party
import pytest

# Local
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.test_helpers.helpers import (
    DummyController,
    config_detail_dict,
    library_config,
)
from oper8.test_helpers.pwm_helpers import (
    MockedReconcileThread,
    heartbeat_file,
    make_managed_object,
    read_heartbeat_file,
)
from oper8.utils import merge_configs
from oper8.watch_manager import PythonWatchManager, WatchManagerBase
from oper8.watch_manager.python_watch_manager.threads import watch
from oper8.watch_manager.python_watch_manager.threads.reconcile import ReconcileThread

## Helpers #####################################################################


class DummyPythonWatchManagerController(DummyController):
    version = "v1"


class SecondDummyPythonWatchManagerController(DummyController):
    version = "v2"


@pytest.fixture(autouse=True)
def reset_watch_base():
    yield
    watch.watch_threads = {}
    WatchManagerBase._ALL_WATCHES = {}


@contextmanager
def mock_reconcile_thread(deploy_manager):
    mock_reconcile_thread = MockedReconcileThread(deploy_manager=deploy_manager)
    ReconcileThread._instance = mock_reconcile_thread
    yield
    ReconcileThread._instance = None


@pytest.mark.timeout(8)
def test_python_watch_manager_happy_path():
    dm = DryRunDeployManager()

    with mock_reconcile_thread(dm):
        watch_manager = PythonWatchManager(
            DummyPythonWatchManagerController, deploy_manager=dm
        )
        assert watch_manager.watch()
        assert watch_manager.reconcile_thread.is_alive()

        for controller_watch in watch_manager.controller_watches:
            assert controller_watch.is_alive()

        resource = make_managed_object()
        time.sleep(1)
        dm.deploy([resource.definition])
        time.sleep(2)
        watch_manager.stop()
        assert watch_manager.reconcile_thread.processes_started == 1
        assert not watch_manager.reconcile_thread.is_alive()


@pytest.mark.timeout(8)
def test_python_watch_manager_namespaced():
    dm = DryRunDeployManager()

    with mock_reconcile_thread(dm):
        watch_manager = PythonWatchManager(
            DummyPythonWatchManagerController,
            deploy_manager=dm,
            namespace_list=["test"],
        )
        assert watch_manager.watch()
        assert watch_manager.reconcile_thread.is_alive()

        for controller_watch in watch_manager.controller_watches:
            assert controller_watch.is_alive()
        time.sleep(1)
        resource = make_managed_object()
        dm.deploy([resource.definition])
        time.sleep(2)
        watch_manager.stop()
        assert watch_manager.reconcile_thread.processes_started == 1
        assert not watch_manager.reconcile_thread.is_alive()


@pytest.mark.timeout(8)
def test_python_watch_manager_multiple_deploy():
    dm = DryRunDeployManager()

    with mock_reconcile_thread(dm):
        watch_manager = PythonWatchManager(
            DummyPythonWatchManagerController, deploy_manager=dm
        )
        second_watch_manager = PythonWatchManager(
            SecondDummyPythonWatchManagerController, deploy_manager=dm
        )
        assert watch_manager.watch()
        assert second_watch_manager.watch()
        assert watch_manager.reconcile_thread == second_watch_manager.reconcile_thread
        assert second_watch_manager.reconcile_thread.is_alive()

        for controller_watch in [
            *watch_manager.controller_watches,
            *second_watch_manager.controller_watches,
        ]:
            assert controller_watch.is_alive()

        time.sleep(1)
        resource = make_managed_object()
        dm.deploy([resource.definition])
        resource = make_managed_object(api_version="foo.bar.com/v2")
        dm.deploy([resource.definition])
        time.sleep(2)
        watch_manager.stop()
        assert watch_manager.reconcile_thread.processes_started == 2
        assert not watch_manager.reconcile_thread.is_alive()


def test_python_watch_manager_singleton():
    dm = DryRunDeployManager()
    watch_manager = PythonWatchManager(DummyPythonWatchManagerController, dm)
    second_watch_manager = PythonWatchManager(DummyController, dm)

    assert id(watch_manager.reconcile_thread) == id(
        second_watch_manager.reconcile_thread
    )


def test_python_watch_manager_heartbeat(heartbeat_file):
    """Test that when configured, PWM keeps a heartbeat file"""
    pwm_config = merge_configs(
        copy.deepcopy(config_detail_dict.python_watch_manager),
        {"heartbeat_file": heartbeat_file, "heartbeat_period": "1s"},
    )
    with library_config(python_watch_manager=pwm_config):
        dm = DryRunDeployManager()
        with mock_reconcile_thread(dm):
            watch_manager = PythonWatchManager(DummyPythonWatchManagerController, dm)
            assert watch_manager.heartbeat_thread

            # Start it and wait for a heartbeat
            assert watch_manager.watch()
            assert watch_manager.reconcile_thread.is_alive()
            watch_manager.heartbeat_thread.wait_for_beat()

            # Make sure the heartbeat writes to a file continuously
            hb1 = read_heartbeat_file(heartbeat_file)
            watch_manager.heartbeat_thread.wait_for_beat()
            hb2 = read_heartbeat_file(heartbeat_file)
            assert hb2 > hb1
            watch_manager.stop()



"""Implementation of the Leader-with-Lease LeaderElection"""
# Standard
from datetime import datetime, timedelta, timezone

# Third Party
from dateutil.parser import parse

# First Party
import alog

# Local
from .... import config
from ....exceptions import assert_config
from ..utils import get_operator_namespace, get_pod_name, parse_time_delta
from .base import ThreadedLeaderManagerBase

log = alog.use_channel("LDRLIFE")


class LeaderWithLeaseManager(ThreadedLeaderManagerBase):
    """
    LeaderWithLeaseManager Class implements the "leader-with-lease" operator-sdk
    lock type. This lock creates a lease object with the operator pod as owner and
    constantly re-acquires the lock.
    """

    def __init__(self, deploy_manager):
        """
        Initialize class with lock_name, current namespace, and pod information
        """
        super().__init__(deploy_manager)

        # Gather lock_name, namespace and pod manifest
        self.lock_name = (
            config.operator_name
            if config.operator_name
            else config.python_watch_manager.lock.name
        )
        self.namespace = get_operator_namespace()
        self.lock_identity = get_pod_name()
        assert_config(self.lock_name, "Unable to detect lock name")
        assert_config(self.namespace, "Unable to detect operator namespace")
        assert_config(self.lock_identity, "Unable to detect lock identity")

    def renew_or_acquire(self):
        """
        Renew or acquire lock by checking the current lease status
        """

        # Template out the expected lease. This is edited based on the current
        # lease status
        current_time = datetime.now(timezone.utc)
        lease_resource_version = None
        expected_lease_data = {
            "holderIdentity": self.lock_identity,
            "acquireTime": current_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
            "leaseDurationSeconds": round(
                parse_time_delta(
                    config.python_watch_manager.lock.duration
                ).total_seconds()
            ),
            "leaseTransitions": 1,
            "renewTime": current_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ"),
        }

        # Get current lease
        success, lease_obj = self.deploy_manager.get_object_current_state(
            kind="Lease",
            name=self.lock_name,
            namespace=self.namespace,
            api_version="coordination.k8s.io/v1",
        )
        if not success:
            log.warning("Unable to fetch lease %s/%s", self.namespace, self.lock_name)

        # If lease exists then verify current holder is valid or update the expected
        # lease with the proper values
        if lease_obj and lease_obj.get("spec"):
            log.debug2(
                "Lease object %s already exists, checking holder", self.lock_name
            )

            lease_resource_version = lease_obj.get("metadata", {}).get(
                "resourceVersion"
            )
            lease_spec = lease_obj.get("spec")
            lock_holder = lease_spec.get("holderIdentity")

            if lock_holder != self.lock_identity:
                renew_time = parse(lease_spec.get("renewTime"))
                lease_duration = timedelta(
                    seconds=lease_spec.get("leaseDurationSeconds")
                )

                # If the renew+lease is after the current time than the other
                # lease holder is still valid
                if (renew_time + lease_duration) > current_time:
                    self.release_lock()
                    return

                log.info("Taking leadership from %s", lock_holder)
                # Increment leaseTransitions as we're taking ownership
                expected_lease_data["leaseTransitions"] = (
                    lease_spec.get("leaseTransitions", 1) + 1
                )

            # If we're the current holder than keep the current acquire time
            else:
                log.debug2(
                    "Lease object already owned. Reusing acquireTime and transitions"
                )
                expected_lease_data["acquireTime"] = lease_spec.get("acquireTime")
                expected_lease_data["leaseTransitions"] = lease_spec.get(
                    "leaseTransitions"
                )

        # Create or update the lease obj
        lease_resource = {
            "kind": "Lease",
            "apiVersion": "coordination.k8s.io/v1",
            "metadata": {
                "name": self.lock_name,
                "namespace": self.namespace,
            },
            "spec": expected_lease_data,
        }
        if lease_resource_version:
            lease_resource["metadata"]["resourceVersion"] = lease_resource_version

        success, _ = self.deploy_manager.deploy(
            [lease_resource], manage_owner_references=False
        )
        if not success:
            log.warning("Unable to acquire leadership lock")
            self.release_lock()
        else:
            self.acquire_lock()



"""
Tests for the FilterManager
"""
# Third Party
import pytest

# First Party
import alog

# Local
from oper8.deploy_manager.kube_event import KubeEventType
from oper8.test_helpers.pwm_helpers import make_managed_object
from oper8.watch_manager.python_watch_manager.filters.filters import (
    CreationDeletionFilter,
    DisableFilter,
    EnableFilter,
    NoGenerationFilter,
)
from oper8.watch_manager.python_watch_manager.filters.manager import (
    AndFilter,
    FilterManager,
    OrFilter,
)

## Helpers #####################################################################


def test_manager_happy_path():
    resource = make_managed_object(generation=None)
    filter = OrFilter(CreationDeletionFilter, NoGenerationFilter)
    fm = FilterManager(filter, resource)

    assert fm.update_and_test(resource, KubeEventType.ADDED)
    updated_resource = make_managed_object(spec={"modified": "value"})
    assert fm.update_and_test(updated_resource, KubeEventType.MODIFIED)
    assert not fm.update_and_test(updated_resource, KubeEventType.MODIFIED)


@pytest.mark.parametrize(
    ["filters", "result"],
    [
        [EnableFilter, True],
        [AndFilter(EnableFilter, DisableFilter), False],
        [OrFilter(EnableFilter, DisableFilter), True],
        [AndFilter(OrFilter(EnableFilter, DisableFilter), EnableFilter), True],
        [AndFilter(OrFilter(DisableFilter, DisableFilter), EnableFilter), False],
    ],
)
def test_manager_conditionals(filters, result):
    resource = make_managed_object()
    fm = FilterManager(filters, resource)
    assert fm.test(resource, KubeEventType.ADDED) == result


def test_manager_to_info():
    filters = OrFilter(CreationDeletionFilter, NoGenerationFilter)

    filter_info = FilterManager.to_info(filters)
    round_tripped_filters = FilterManager.from_info(filter_info)

    assert round_tripped_filters == filters



"""
Tests for the LeaderForLifeManager
"""
# Standard
import time

# Third Party
import pytest

# Local
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.exceptions import ConfigError
from oper8.test_helpers.helpers import library_config
from oper8.test_helpers.pwm_helpers import (
    MockedLeaderForLifeManager,
    make_managed_object,
    make_ownerref,
)

## Helpers #####################################################################


@pytest.mark.timeout(5)
def test_life_happy_path():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([resource.definition])
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "namespace": resource.namespace,
            }
        },
    ):
        lm = MockedLeaderForLifeManager(deploy_manager=dm)
        assert lm.acquire()
        assert lm.is_leader()
        assert lm.acquire_resource(None)
        lm.release()
        assert not lm.is_leader()


@pytest.mark.timeout(5)
def test_life_already_owner():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([resource.definition])
    existing_cm = make_managed_object(
        kind="ConfigMap",
        api_version="v1",
        name="test_lock",
        namespace=resource.namespace,
        owner_refs=[make_ownerref(resource.definition)],
    )
    dm.deploy([existing_cm.definition])
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "namespace": resource.namespace,
            }
        },
    ):
        lm = MockedLeaderForLifeManager(deploy_manager=dm)
        assert lm.acquire()
        assert lm.is_leader()
        assert lm.acquire_resource(None)
        lm.release()
        assert not lm.is_leader()


@pytest.mark.timeout(5)
def test_life_unknown_pod():
    dm = DryRunDeployManager()
    resource = make_managed_object(kind="Pod", api_version="v1")
    with library_config(
        pod_name=resource.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "namespace": resource.namespace,
            }
        },
    ):
        with pytest.raises(ConfigError):
            lm = MockedLeaderForLifeManager(deploy_manager=dm)
            lm.acquire()


@pytest.mark.timeout(5)
def test_life_bad_config_map():
    dm = DryRunDeployManager()
    original_owner = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([original_owner.definition])
    current_pod = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([current_pod.definition])
    existing_cm = make_managed_object(
        kind="ConfigMap",
        api_version="v1",
        name="test_lock",
        namespace=original_owner.namespace,
        owner_refs=[
            make_ownerref(original_owner.definition),
            make_ownerref(current_pod.definition),
        ],
    )
    dm.deploy([existing_cm.definition])

    with library_config(
        pod_name=current_pod.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "namespace": current_pod.namespace,
            }
        },
    ):
        lm = MockedLeaderForLifeManager(deploy_manager=dm)
        lm.run_renew_or_acquire()
        assert not lm.is_leader()


@pytest.mark.timeout(5)
def test_life_invalid_config():
    dm = DryRunDeployManager()
    with library_config(
        python_watch_manager={"lock": {"name": "test_lock", "poll_time": "1s"}}
    ):
        with pytest.raises(ConfigError):
            lm = MockedLeaderForLifeManager(deploy_manager=dm)
            lm.acquire()


@pytest.mark.timeout(5)
def test_life_not_owner():
    dm = DryRunDeployManager()
    original_owner = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([original_owner.definition])
    existing_cm = make_managed_object(
        kind="ConfigMap",
        api_version="v1",
        name="test_lock",
        namespace=original_owner.namespace,
        owner_refs=[make_ownerref(original_owner.definition)],
    )
    dm.deploy([existing_cm.definition])
    current_pod = make_managed_object(kind="Pod", api_version="v1")
    dm.deploy([current_pod.definition])
    with library_config(
        pod_name=current_pod.name,
        python_watch_manager={
            "lock": {
                "name": "test_lock",
                "poll_time": "1s",
                "namespace": current_pod.namespace,
            }
        },
    ):
        lm = MockedLeaderForLifeManager(deploy_manager=dm)
        lm.run_renew_or_acquire()
        assert not lm.is_leader()


@pytest.mark.timeout(5)
def test_life_competition():
    dm = DryRunDeployManager(strict_resource_version=True)
    first_operator = make_managed_object(kind="Pod", api_version="v1", name="first")
    dm.deploy([first_operator.definition])
    with library_config(
        pod_name=first_operator.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "namespace": first_operator.namespace,
            }
        },
    ):
        first_manager = MockedLeaderForLifeManager(deploy_manager=dm)

    second_operator = make_managed_object(kind="Pod", api_version="v1", name="second")
    dm.deploy([second_operator.definition])
    with library_config(
        pod_name=second_operator.name,
        python_watch_manager={
            "lock": {
                "name": "test",
                "poll_time": "1s",
                "namespace": second_operator.namespace,
            }
        },
    ):
        second_manager = MockedLeaderForLifeManager(deploy_manager=dm)

    first_manager.leadership_thread.start()
    # Add slight delay as DryRunDeployManager doesn't handle resource to old/out of date exceptions
    time.sleep(0.1)
    second_manager.leadership_thread.start()
    time.sleep(1)
    # Assert only one of them gained leadership
    assert first_manager.is_leader() ^ second_manager.is_leader()
    first_manager.release()
    second_manager.release()



"""
This module holds common helper functions for making testing easy
"""

# Standard
from contextlib import contextmanager
from datetime import datetime
from functools import lru_cache
from unittest import mock
import copy
import hashlib
import inspect
import json
import os
import pathlib
import subprocess
import tempfile
import uuid

# Third Party
from kubernetes.dynamic.client import DynamicClient
import pytest

# First Party
import aconfig
import alog

# Local
from oper8 import Component, Controller, controller
from oper8.cmd.run_operator_cmd import RunOperatorCmd
from oper8.config import library_config as config_detail_dict
from oper8.deploy_manager.dry_run_deploy_manager import DryRunDeployManager
from oper8.deploy_manager.openshift_deploy_manager import OpenshiftDeployManager
from oper8.session import Session
from oper8.test_helpers.kub_mock import MockKubClient
from oper8.utils import merge_configs

log = alog.use_channel("TEST")


def configure_logging():
    alog.configure(
        os.environ.get("LOG_LEVEL", "off"),
        os.environ.get("LOG_FILTERS", ""),
        formatter="json"
        if os.environ.get("LOG_JSON", "").lower() == "true"
        else "pretty",
        thread_id=os.environ.get("LOG_THREAD_ID", "").lower() == "true",
    )


configure_logging()

TEST_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

TEST_INSTANCE_NAME = "test_instance"
TEST_INSTANCE_UID = "12345678-1234-1234-1234-123456789012"
TEST_NAMESPACE = "test"
SOME_OTHER_NAMESPACE = "somewhere"

PULL_SECRET_NAME = "REPLACE_PULL_SECRET"


def setup_cr(
    kind="Widget",
    api_version="foo.bar.com/v123",
    deploy_config=None,
    version="1.2.3",
    name=TEST_INSTANCE_NAME,
    namespace=TEST_NAMESPACE,
    **kwargs,
):
    deploy_config = deploy_config or {}
    cr_dict = kwargs or {}
    cr_dict.setdefault("kind", kind)
    cr_dict.setdefault("apiVersion", api_version)
    cr_dict.setdefault("metadata", {}).setdefault("name", name)
    cr_dict.setdefault("metadata", {}).setdefault("namespace", namespace)
    cr_dict.setdefault("metadata", {}).setdefault("uid", TEST_INSTANCE_UID)
    cr_dict.setdefault("spec", {}).update(copy.deepcopy(deploy_config))
    cr_dict["spec"].setdefault("version", version)
    return aconfig.Config(cr_dict)


def setup_session(
    version="1.2.3",
    app_config=None,
    deploy_config=None,
    full_cr=None,
    deploy_manager=None,
    namespace=TEST_NAMESPACE,
    deploy_initial_cr=True,
    **kwargs,
):
    app_config = app_config or aconfig.Config({}, override_env_vars=False)
    deploy_config = deploy_config or aconfig.Config({}, override_env_vars=False)
    full_cr = full_cr or setup_cr(
        deploy_config=deploy_config, version=version, namespace=namespace
    )
    if not deploy_manager:
        deploy_manager = (
            MockDeployManager(resources=[full_cr])
            if deploy_initial_cr
            else MockDeployManager()
        )

    return Session(
        reconciliation_id=str(uuid.uuid4()),
        cr_manifest=full_cr,
        config=app_config,
        deploy_manager=deploy_manager,
        **kwargs,
    )


@contextmanager
def setup_session_ctx(*args, **kwargs):
    """Context manager wrapper around setup_session. This simplifies the porting
    process from WA and really provides no functional benefit.
    """
    yield setup_session(*args, **kwargs)


@contextmanager
def library_config(**config_overrides):
    """This context manager sets library config values temporarily and reverts
    them on completion
    """
    # Override the configs and hang onto the old values
    old_vals = {}
    for key, val in config_overrides.items():
        if key in config_detail_dict:
            old_vals[key] = config_detail_dict[key]
        config_detail_dict[key] = val

    # Yield to the context
    yield

    # Revert to the old values
    for key in config_overrides:
        if key in old_vals:
            config_detail_dict[key] = old_vals[key]
        else:
            del config_detail_dict[key]


def get_failable_method(fail_flag, method, failure_return=False):
    log.debug4(
        "Setting up failable mock of [%s] with fail flag: %s", str(method), fail_flag
    )

    def failable_method(*args, **kwargs):
        log.debug4(
            "Running failable mock of [%s] with fail flag: %s", str(method), fail_flag
        )
        if isinstance(fail_flag, Exception) or (
            inspect.isclass(fail_flag) and issubclass(fail_flag, Exception)
        ):
            log.debug4("Raising in failable mock")
            raise fail_flag
        elif callable(fail_flag):
            log.debug4("Calling callable fail flag")
            res = fail_flag()
            if res is not None:
                return res
        elif fail_flag == "assert":
            log.debug4("Asserting in failable mock")
            raise AssertionError(f"You told me to fail {method}!")
        elif fail_flag:
            log.debug4("Returning %s", failure_return)
            return failure_return
        log.debug4("Passing through (%s, **%s)", args, kwargs)
        res = method(*args, **kwargs)
        log.debug4("Passthrough res: %s", res)
        return res

    return failable_method


class FailOnce:
    """Helper callable that will fail once on the N'th call"""

    def __init__(self, fail_val, fail_number=1):
        self.call_count = 0
        self.fail_number = fail_number
        self.fail_val = fail_val

    def __call__(self, *_, **__):
        self.call_count += 1
        if self.call_count == self.fail_number:
            log.debug("Failing on call %d with %s", self.call_count, self.fail_val)
            if isinstance(self.fail_val, type) and issubclass(self.fail_val, Exception):
                raise self.fail_val("Raising!")
            return self.fail_val
        log.debug("Not failing on call %d", self.call_count)
        return


class MockDeployManager(DryRunDeployManager):
    """The MockDeployManager wraps a standard DryRunDeployManager and adds
    configuration options to simulate failures in each of its operations.
    """

    def __init__(
        self,
        deploy_fail=False,
        deploy_raise=False,
        disable_fail=False,
        disable_raise=False,
        get_state_fail=False,
        get_state_raise=False,
        set_status_fail=False,
        set_status_raise=False,
        auto_enable=True,
        resources=None,
        resource_dir=None,
        **kwargs,
    ):
        """This DeployManager can be configured to have various failure cases
        and will mock the state of the cluster so that get_object_current_state
        will pull its information from the local dict.
        """

        # Add apiVersion to resources that are missing it, then initialize the
        # dry run manager

        resources = resources or []
        # Parse pre-populated resources if needed
        resources = resources + (RunOperatorCmd._parse_resource_dir(resource_dir))

        for resource in resources:
            resource.setdefault("apiVersion", "v1")
        super().__init__(resources, **kwargs)

        self.deploy_fail = "assert" if deploy_raise else deploy_fail
        self.disable_fail = "assert" if disable_raise else disable_fail
        self.get_state_fail = "assert" if get_state_raise else get_state_fail
        self.set_status_fail = "assert" if set_status_raise else set_status_fail

        # If auto-enabling, turn the mocks on now
        if auto_enable:
            self.enable_mocks()

    #######################
    ## Helpers for Tests ##
    #######################

    def enable_mocks(self):
        """Turn the mocks on"""
        self.deploy = mock.Mock(
            side_effect=get_failable_method(
                self.deploy_fail, super().deploy, (False, False)
            )
        )
        self.disable = mock.Mock(
            side_effect=get_failable_method(
                self.disable_fail, super().disable, (False, False)
            )
        )
        self.get_object_current_state = mock.Mock(
            side_effect=get_failable_method(
                self.get_state_fail, super().get_object_current_state, (False, None)
            )
        )
        self.set_status = mock.Mock(
            side_effect=get_failable_method(
                self.set_status_fail, super().set_status, (False, False)
            )
        )

    def get_obj(self, kind, name, namespace=None, api_version=None):
        return self.get_object_current_state(kind, name, namespace, api_version)[1]

    def has_obj(self, *args, **kwargs):
        return self.get_obj(*args, **kwargs) is not None


default_dummy_metadata = {"labels": {"foo": "bar"}}


dummy_instance_spec = {
    "kind": "unicorn",
    "metadata": {"name": "instance", "namespace": TEST_NAMESPACE},
    "labels": {"label": "instance-label"},
}


def make_patch(
    patch_type,
    body,
    name="test",
    target=None,
    namespace=TEST_NAMESPACE,
    api_version="org.oper8/v1",
    kind="TemporaryPatch",
):
    """Make a sample TemporaryPatch resource body"""
    target = target or {}
    patch_obj = {
        "apiVersion": api_version,
        "kind": kind,
        "metadata": {"name": name},
        "spec": {
            "apiVersion": target.get("apiVersion", "fake"),
            "kind": target.get("kind", "fake"),
            "name": target.get("metadata", {}).get("name", "fake"),
            "patchType": patch_type,
            "patch": body,
        },
    }
    if namespace is not None:
        patch_obj["metadata"]["namespace"] = namespace
    return aconfig.Config(
        patch_obj,
        override_env_vars=False,
    )


## DummyComponent ##############################################################


class DummyComponentBase(Component):
    """This base class provides all of the common functionality for
    DummyComponent and DummyLegacyComponent
    """

    def __init__(
        self,
        name=None,
        session=None,
        api_objects=None,
        api_object_deps=None,
        render_chart_fail=False,
        deploy_fail=False,
        disable_fail=False,
        verify_fail=False,
        build_chart_fail=False,
        disabled=False,
        **kwargs,
    ):
        # Hang onto config inputs
        self.api_objects = api_objects or []
        self.api_object_deps = api_object_deps or {}

        # Mock passthroughs to the base class
        self.render_chart_fail = render_chart_fail
        self.render_chart = mock.Mock(
            side_effect=get_failable_method(self.render_chart_fail, self.render_chart)
        )
        self.deploy_fail = deploy_fail
        self.deploy = mock.Mock(
            side_effect=get_failable_method(self.deploy_fail, super().deploy)
        )
        self.disable_fail = disable_fail
        self.disable = mock.Mock(
            side_effect=get_failable_method(self.disable_fail, super().disable)
        )
        self.verify_fail = verify_fail
        self.verify = mock.Mock(
            side_effect=get_failable_method(self.verify_fail, super().verify)
        )
        self.build_chart_fail = build_chart_fail
        self.build_chart = mock.Mock(
            side_effect=get_failable_method(self.build_chart_fail, self.build_chart)
        )

        # Initialize Component
        if name is None:
            super().__init__(session=session, disabled=disabled)
        else:
            super().__init__(name=name, session=session, disabled=disabled)

    @alog.logged_function(log.debug2)
    def _add_resources(self, scope, session):
        """This will be called in both implementations in their respective
        places
        """
        api_objs = self._gather_dummy_resources(scope, session)

        # Add dependencies between objects
        for downstream_name, upstreams in self.api_object_deps.items():
            assert downstream_name in api_objs, "Bad test config"
            downstream = api_objs[downstream_name]
            for upstream_name in upstreams:
                assert upstream_name in api_objs, "Bad test config"
                upstream = api_objs[upstream_name]
                downstream.add_dependency(upstream)

    def _gather_dummy_resources(self, scope, session):
        api_objs = {}
        for api_obj in self.api_objects:
            log.debug3("Creating api object: %s", api_obj)

            # Create api object from tuple or generate one if callable
            if isinstance(api_obj, tuple):
                object_name, object_def = api_obj

                object_def = merge_configs(
                    {
                        "apiVersion": "v1",
                        "metadata": {"name": object_name, "namespace": TEST_NAMESPACE},
                    },
                    object_def,
                )
            elif isinstance(api_obj, dict):
                object_def = api_obj
                object_name = api_obj.get("metadata", {}).get("name")
            else:
                object_def = api_obj(self, session)
                object_name = api_obj.name

            resource_node = self.add_resource(object_name, object_def)
            if resource_node is not None:
                api_objs[resource_node.get_name()] = resource_node
        return api_objs

    def get_rendered_configs(self):
        configs = []
        for obj in self.managed_objects:
            configs.append(aconfig.Config(obj.definition))
        return configs


@pytest.fixture(autouse=True)
def version_safe_md5():
    real_md5 = hashlib.md5

    UNDEFINED = "__undefined__"

    def md5_wrapper(first_arg=UNDEFINED, *_, **__):
        if first_arg is UNDEFINED:
            return real_md5()
        return real_md5(first_arg)

    with mock.patch.object(hashlib, "md5", md5_wrapper):
        yield


class MockedOpenshiftDeployManager(OpenshiftDeployManager):
    """Override class that uses the mocked client"""

    def __init__(self, manage_ansible_status=False, owner_cr=None, *args, **kwargs):
        self._mock_args = args
        self._mock_kwargs = kwargs
        super().__init__(manage_ansible_status, owner_cr)

    def _setup_client(self):
        mock_client = MockKubClient(*self._mock_args, **self._mock_kwargs)
        return DynamicClient(mock_client)


class DummyNodeComponent(DummyComponentBase):
    """
    Configurable dummy component which will create an abritrary set of
    resource node instances.
    """

    def __init__(self, session, *args, **kwargs):
        """Construct with the additional option to fail build_chart"""
        super().__init__(*args, session=session, **kwargs)
        self._add_resources(self, session)


class MockComponent(DummyNodeComponent):
    """Dummy component with a valid mock name"""

    name = "mock"


## DummyController #############################################################


class DummyController(Controller):
    """Configurable implementation of a controller that can be used in unit
    tests to simulate Controller behavior
    """

    ##################
    ## Construction ##
    ##################

    group = "foo.bar.com"
    version = "v42"
    kind = "Foo"

    def __init__(
        self,
        components=None,
        after_deploy_fail=False,
        after_verify_fail=False,
        setup_components_fail=False,
        finalize_components_fail=False,
        should_requeue_fail=False,
        component_type=DummyNodeComponent,
        **kwargs,
    ):
        # Initialize parent
        super().__init__(**kwargs)

        # Set up objects that this controller will manage directly
        self.component_specs = components or []
        self.component_type = component_type

        # Set up mocks
        self.after_deploy_fail = after_deploy_fail
        self.after_verify_fail = after_verify_fail
        self.setup_components_fail = setup_components_fail
        self.finalize_components_fail = finalize_components_fail
        self.should_requeue_fail = should_requeue_fail
        self.after_deploy = mock.Mock(
            side_effect=get_failable_method(
                self.after_deploy_fail, super().after_deploy
            )
        )
        self.after_verify = mock.Mock(
            side_effect=get_failable_method(
                self.after_verify_fail, super().after_verify
            )
        )
        self.setup_components = mock.Mock(
            side_effect=get_failable_method(
                self.setup_components_fail, self.setup_components
            )
        )
        self.finalize_components = mock.Mock(
            side_effect=get_failable_method(
                self.finalize_components_fail, self.finalize_components
            )
        )
        self.should_requeue = mock.Mock(
            side_effect=get_failable_method(
                self.should_requeue_fail, super().should_requeue
            )
        )

    ##############################
    ## Interface Implementation ##
    ##############################

    def setup_components(self, session: Session):
        """Set up the components based on the component specs passed in"""

        # Add the components
        for component in self.component_specs:
            name = component["name"]
            log.debug2("Adding component %s (kwargs: %s)", name, component)
            comp = self._make_dummy_component(
                session=session,
                **component,
            )
            log.debug2("Component name: %s", comp.name)
            log.debug2(
                "Components in session [%s]: %s",
                session.id,
                [
                    comp.name
                    for comp in session.get_components()
                    + session.get_components(disabled=True)
                ],
            )

        # Add the dependencies after the nodes (so that we can look up by name)
        component_map = {
            comp.name: comp
            for comp in session.get_components() + session.get_components(disabled=True)
        }
        for component in self.component_specs:
            comp = component_map[component["name"]]
            upstreams = component.get("upstreams", [])
            for upstream in upstreams:
                session.add_component_dependency(comp, upstream)

        # Hang onto the components so that they can be checked
        self.components = component_map

    ############################
    ## Implementation Details ##
    ############################

    def _make_dummy_component(self, name="dummy", session=None, **kwargs):
        """This helper wraps any DummyComponent class so that the name class
        attribute is not overwritten by the next instance.
        """

        class WrappedDummyComponent(self.component_type):
            pass

        WrappedDummyComponent.name = name
        return WrappedDummyComponent(session=session, **kwargs)


@controller(
    group="unit.test.com",
    version="v42",
    kind="MockTopApp",
)
class MockTopApp(Controller):
    """Mock implementation of a top-level Controller to allow subsystems to be
    tested as "children"
    """

    def __init__(self, config_defaults=None, component_types=None):
        super().__init__(config_defaults=config_defaults)
        self.component_types = component_types or []

    def setup_components(self, session: Session):
        for component_type in self.component_types:
            component_type(session=session)

    def do_rollout(self, session):
        try:
            return self.run_reconcile(session)
        except Exception as err:
            log.debug("Caught error in rollout: %s", err, exc_info=True)


## Ansible Tests ###############################################################


class ModuleExit(Exception):
    """Exception we'll use to break out when sys.exit was called"""


class TestRecorder:
    # This is not a test class!
    __test__ = False

    CONSTRUCTED = "constructed"
    SETUP_COMPONENTS = "setup_components"
    FINALIZE_COMPONENTS = "finalize_components"
    RECONCILE = "reconcile"
    EXITED = "exited"

    def __init__(self, raise_on_success=True):
        self.retcode = None
        self.events = []
        self.event_details = {}
        self.raise_on_success = raise_on_success

    def set_exit_code(self, code=0):
        self.events.append(self.EXITED)
        self.retcode = code
        if code != 0 or self.raise_on_success:
            raise ModuleExit()

    def add_event(self, event, details=None):
        self.events.append(event)
        if details is not None:
            self.event_details[event] = details


@contextmanager
def mock_config_file(config_object):
    """Yuck! Ansible makes it tough to actually inject parameters in since it
    expects that modules will only be run by its parent runner.
    """
    # Third Party
    import ansible.module_utils.basic

    ansible.module_utils.basic._ANSIBLE_ARGS = json.dumps(config_object).encode("utf-8")
    yield
    ansible.module_utils.basic._ANSIBLE_ARGS = None


@contextmanager
def mock_sys_exit(recorder):
    with mock.patch("sys.exit", recorder.set_exit_code):
        yield


## General Helpers #############################################################


@lru_cache(maxsize=1)
def default_branch_name() -> str:
    """Helper to get the current git context's default branch name"""
    try:
        return (
            subprocess.run(
                "git config --get init.defaultBranch".split(),
                check=True,
                stdout=subprocess.PIPE,
            )
            .stdout.decode("utf-8")
            .strip()
        )
    except subprocess.CalledProcessError:
        return "master"


def setup_vcs_project(
    components: list,
    directory: str,
    controller_name="Controller",
    module_dir="src",
    module_name="test_module",
    versions=None,
):
    # Make the parent directory and module directory
    versions = versions or ["1.2.3"]
    parent_path = pathlib.Path(directory)
    parent_path.mkdir(exist_ok=True)
    module_path = parent_path / module_dir / module_name
    module_path.mkdir(parents=True)

    # Open and template controller file
    template_file = pathlib.Path(TEST_DATA_DIR) / "controller.template"
    with open(template_file) as controller_template_file:
        controller_template = controller_template_file.read()

    # Create controller file from template
    with open(module_path / "controller.py", "w") as controller_file:
        controller_file.write(
            controller_template.format(
                controller_name=controller_name,
                components=str(components),
            )
        )

    # Create import file
    with open(module_path / "__init__.py", "w") as import_file:
        import_file.write(f"from .controller import {controller_name}")

    config_path = pathlib.Path.home() / ".gitconfig"
    if not config_path.exists():
        config_path.touch(exist_ok=True)

    # Initialize git repo in parent directory
    subprocess.run(["git", "init", directory], check=True)
    subprocess.run(["git", "-C", directory, "add", "."], check=True)
    subprocess.run(
        [
            "git",
            "-c",
            "user.name='Oper8'",
            "-c",
            "user.email='my@email.org'",
            "-C",
            directory,
            "commit",
            "-m",
            "Initial Commit",
        ],
        check=True,
    )

    # For each version create a branch and test file. This ensures each version has a different
    # commit hash
    for version in versions:
        subprocess.run(["git", "-C", directory, "checkout", "-b", version], check=True)
        (parent_path / version).touch(exist_ok=True)
        subprocess.run(["git", "-C", directory, "add", "."], check=True)
        subprocess.run(
            [
                "git",
                "-c",
                "user.name='Oper8'",
                "-c",
                "user.email='my@email.org'",
                "-C",
                directory,
                "commit",
                "-m",
                "Annother Commit",
            ],
            check=True,
        )
        subprocess.run(["git", "-C", directory, "tag", version], check=True)
        subprocess.run(
            ["git", "-C", directory, "checkout", default_branch_name()], check=True
        )


@pytest.fixture()
def vcs_project():
    with tempfile.TemporaryDirectory() as vcs_directory:
        setup_vcs_project(
            components=[],
            directory=vcs_directory,
            module_name="test_module",
            versions=["1.2.3"],
            module_dir="src",
        )

        yield vcs_directory


@contextmanager
def maybe_temp_dir():
    working_dir = os.environ.get("WORKING_DIR")
    if working_dir:
        working_dir = f"{working_dir}.{datetime.now().timestamp()}"
        os.makedirs(working_dir, exist_ok=True)
        yield working_dir
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield temp_dir


def deep_merge(a, b):
    """NOTE: This should really be eliminated in favor of just using
    merge_configs
    """
    return merge_configs(a, b)



"""Log handler helper class"""

# Standard
from logging import Formatter, LogRecord
from logging.handlers import QueueHandler
import copy

# First Party
import alog

# Local
from ....managed_object import ManagedObject

log = alog.use_channel("LOG-HANDLER")

# Forward declaration of a queue for any type
QUEUE_TYPE = "Queue[Any]"


class LogQueueHandler(QueueHandler):
    """
    Log Handler class to collect messages from a child processes and pass
    them to the root process via a multiprocess queue
    """

    def __init__(self, queue: QUEUE_TYPE, manifest: ManagedObject = None):
        """Initialize the queue handler and instance variables

        Args:
            queue: "Queue[Any]"
                The queue to pass messages to
            manifest: ManagedObject
                The manifest of the current process. This is only used if it can't find
                the resource on the current formatter
        """
        super().__init__(queue)
        self.manifest = manifest

    def prepare(self, record: LogRecord) -> LogRecord:
        """Prep a record for pickling before sending it to the queue

        Args:
            record: LogRecord
                The record to be prepared

        Returns:
            prepared_record: LogRecord
                The prepared record ready to be pickled
        """

        # Duplicate record to preserve other handlers
        record = copy.copy(record)

        # get the currently used formatter
        formatter = self.formatter if self.formatter else Formatter()

        # Exceptions can't always be pickled so manually process
        # the record but remove the exc_info This retains the
        # the processed exc_txt but allows the parent process to reformat
        # the message
        if record.exc_info:
            record.exc_text = formatter.formatException(record.exc_info)
            record.exc_info = None

        # In case there are exceptions/unpicklable objects in the logging
        # args then manually compute the message. After computing clear the
        # message&args values to allow the parent process to reformat the
        # record
        record.msg = record.getMessage()
        record.args = []

        # Take the manifest from the current formatter and pass it back up
        resource = {}
        if hasattr(formatter, "manifest"):
            resource = formatter.manifest
        elif self.manifest:
            resource = self.manifest

        # Only copy required resource keys to the record
        resource_metadata = resource.get("metadata", {})
        record.resource = {
            "kind": resource.get("kind"),
            "apiVersion": resource.get("apiVersion"),
            "metadata": {
                "name": resource_metadata.get("name"),
                "namespace": resource_metadata.get("namespace"),
                "resourceVersion": resource_metadata.get("resourceVersion"),
            },
        }

        return record



"""
Test the patch_strategic_merge implementation
"""

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.patch_strategic_merge import patch_strategic_merge
from oper8.test_helpers.helpers import configure_logging

configure_logging()

log = alog.use_channel("TEST")

## Helpers #####################################################################


def sample_foo(base):
    base.setdefault("kind", "Foo")
    base.setdefault("apiVersion", "v1")
    base.setdefault("metadata", {}).setdefault("name", "foo")
    return base


def sample_pod(containers=None):
    return {
        "kind": "Pod",
        "apiVersion": "v1",
        "metadata": {"name": "foo"},
        "spec": {"containers": containers or []},
    }


def pod_psm_body(containers=None):
    out = {}
    out.setdefault("spec", {})["containers"] = containers or []
    return out


## Tests #######################################################################

################
## Happy Path ##
################


def test_apply_patch():
    """Test that a simple patchStrategicMerge patch applies cleanly"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = pod_psm_body(
        [
            {"name": "foo", "restartPolicy": "Always"},
            {"name": "bar", "image": "bar"},
        ]
    )
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 2
    assert res["spec"]["containers"][0] == {
        "name": "foo",
        "image": "foo",
        "restartPolicy": "Always",
    }
    assert res["spec"]["containers"][1] == {"name": "bar", "image": "bar"}


def test_update_key():
    """Test that a key can be updated in an object"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = pod_psm_body(
        [
            {"name": "foo", "restartPolicy": "OnFailure"},
        ]
    )
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 1
    assert res["spec"]["containers"][0] == {
        "name": "foo",
        "image": "foo",
        "restartPolicy": "OnFailure",
    }


def test_delete_key():
    """Test that a key can be deleted from an object"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = pod_psm_body(
        [
            {"name": "foo", "restartPolicy": None},
        ]
    )
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 1
    assert res["spec"]["containers"][0] == {"name": "foo", "image": "foo"}


def test_delete_dict_list_item():
    """Test that an element can be deleted from a list of dicts"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = pod_psm_body(
        [
            {"name": "foo", "$patch": "delete"},
        ]
    )
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 0


def test_add_dict_list_item():
    """Test that an element can be added to a list of dicts"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    patch = pod_psm_body(
        [
            {"name": "bar", "restartPolicy": "Maybe"},
        ]
    )
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 2


def test_delete_primitive_list_items():
    """Test that an element can be deleted from a list of primitives"""
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = {"$deleteFromPrimitiveList/key": [2, 4]}
    res = patch_strategic_merge(obj, patch)
    assert res["key"] == [1, 3]


def test_list_element_replace():
    """Test that an element in a list can be replaced without merging"""
    obj = sample_pod([{"name": "foo", "image": "foo", "restartPolicy": "Always"}])
    replaced_foo = {
        "name": "foo",
        "$patch": "replace",
        "image": "foo:latest",
        "foo": "bar",
    }
    patch = pod_psm_body([replaced_foo])
    res = patch_strategic_merge(obj, patch)
    assert len(res["spec"]["containers"]) == 1
    del replaced_foo["$patch"]
    assert res["spec"]["containers"][0] == replaced_foo


def test_list_no_merge_key_replace():
    """Test that for a list with no merge key, it is directly replaced"""
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = {"key": [2, 4]}
    res = patch_strategic_merge(obj, patch)
    assert res["key"] == [2, 4]


def test_list_custom_merge_patch_keys():
    """Test that list merging is performed correctly when a custom set of
    merge_patch_keys is given
    """
    obj = sample_foo(
        {
            "key": [
                {"name": "one", "nestedKey": "foo", "baz": "bat"},
                {"name": "two", "nestedKey": "bar", "baz": "bat"},
            ]
        }
    )
    patch = {"key": [{"name": "two", "nestedKey": "OVERRIDE"}]}
    res = patch_strategic_merge(obj, patch, merge_patch_keys={"Foo.key": "name"})
    assert res["key"] == [
        {"name": "one", "nestedKey": "foo", "baz": "bat"},
        {"name": "two", "nestedKey": "OVERRIDE", "baz": "bat"},
    ]


#################
## Error Cases ##
#################


def test_bad_directive():
    """Test that an unknown directive throws an error"""
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = pod_psm_body([{"name": "foo", "$patch": "flipflop"}])
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_missing_merge_key_patch():
    """Test that an element in a patch list that is missing the merge key
    causes an error
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = pod_psm_body([{"foo": "bar"}])
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_missing_merge_key_current():
    """Test that an element in a current list that is missing the merge key
    causes an error
    """
    obj = sample_pod([{"image": "foo"}])
    patch = pod_psm_body([{"name": "foo", "foo": "bar"}])
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_primitive_delete_not_there():
    """Test that a request to delete a primitive from a list causes an error if
    the element is not present
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = {"$deleteFromPrimitiveList/keyNotThere": [2, 4]}
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_primitive_delete_patch_not_list():
    """Test that a request to delete a primitive from a list where the patch
    value is not a list causes an error
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = {"$deleteFromPrimitiveList/key": 2}
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_primitive_delete_current_not_list():
    """Test that a request to delete a primitive from a list where the current
    value is not a list causes an error
    """
    obj = sample_foo({"key": {"a": 1, "b": 2}})
    patch = {"$deleteFromPrimitiveList/key": ["a"]}
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_primitive_delete_element_not_found():
    """Test that a request to delete a primitive from a list which doesn't
    contain the element raises an error
    """
    obj = sample_foo({"key": [1, 2, 3, 4]})
    patch = {"$deleteFromPrimitiveList/key": [5]}
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)


def test_delete_element_not_found():
    """Test that a `$patch: delete` directive to delete an element from a list
    that isn't found raises an error
    """
    obj = sample_pod([{"name": "foo", "image": "foo"}])
    patch = pod_psm_body([{"name": "bar", "$patch": "delete"}])
    with pytest.raises(ValueError):
        patch_strategic_merge(obj, patch)



"""
This is the main entrypoint command for running the operator
"""
# Standard
from typing import List, Optional, Type
import argparse
import importlib
import os
import signal

# Third Party
import yaml

# First Party
import alog

# Local
from .. import config, watch_manager
from ..controller import Controller
from ..deploy_manager import DryRunDeployManager
from ..exceptions import ConfigError
from .base import CmdBase

log = alog.use_channel("MAIN")


class RunOperatorCmd(CmdBase):
    __doc__ = __doc__

    ## Interface ##

    def add_subparser(
        self,
        subparsers: argparse._SubParsersAction,
    ) -> argparse.ArgumentParser:
        parser = subparsers.add_parser("run", help=__doc__)
        runtime_args = parser.add_argument_group("Runtime Configuration")
        runtime_args.add_argument(
            "--module_name",
            "-m",
            required=True,
            help="The module to import that holds the operator code",
        )
        runtime_args.add_argument(
            "--cr",
            "-c",
            default=None,
            help="(dry run) A CR manifest yaml to apply directly ",
        )
        runtime_args.add_argument(
            "--resource_dir",
            "-r",
            default=None,
            help="(dry run) Path to a directory of yaml files that should exist in the cluster",
        )
        return parser

    def cmd(self, args: argparse.Namespace):
        # Validate args
        assert args.cr is None or (
            config.dry_run and os.path.isfile(args.cr)
        ), "Can only specify --cr with dry run and it must point to a valid file"
        assert args.resource_dir is None or (
            config.dry_run and os.path.isdir(args.resource_dir)
        ), "Can only specify --resource_dir with dry run and it must point to a valid directory"

        # Find all controllers in the operator library
        controller_types = self._get_controller_types(
            args.module_name, args.controller_name
        )

        # Parse pre-populated resources if needed
        resources = self._parse_resource_dir(args.resource_dir)

        # Create the watch managers
        deploy_manager = self._setup_watches(controller_types, resources)

        # Register the signal handler to stop the watches
        def do_stop(*_, **__):  # pragma: no cover
            watch_manager.stop_all()

        signal.signal(signal.SIGINT, do_stop)

        # Run the watch manager
        log.info("Starting Watches")
        watch_manager.start_all()

        # If given, apply the CR directly
        if args.cr:
            log.info("Applying CR [%s]", args.cr)
            with open(args.cr, encoding="utf-8") as handle:
                cr_manifest = yaml.safe_load(handle)
                cr_manifest.setdefault("metadata", {}).setdefault(
                    "namespace", "default"
                )
                log.debug3(cr_manifest)
                deploy_manager.deploy([cr_manifest])

        # All done!
        log.info("SHUTTING DOWN")

    ## Impl ##

    @staticmethod
    def _is_controller_type(attr_val: str):
        """Determine if a given attribute value is a controller type"""
        return (
            isinstance(attr_val, type)
            and issubclass(attr_val, Controller)
            and attr_val is not Controller
        )

    @classmethod
    def _get_controller_types(cls, module_name: str, controller_name=""):
        """Import the operator library and either extract all Controllers,
        or just extract the provided Controller"""
        module = importlib.import_module(module_name)
        log.debug4(dir(module))
        controller_types = []

        if controller_name:
            # Confirm that the class exists and that it is a controller type
            try:
                controller_attr_val = getattr(module, controller_name)
                is_valid_controller = cls._is_controller_type(controller_attr_val)
            except AttributeError:
                is_valid_controller = False

            if is_valid_controller:
                log.debug3("Provided controller, %s, is valid", controller_name)
                controller_types.append(controller_attr_val)
            else:
                raise AttributeError(
                    f"Provided controller, {controller_name}, is invalid"
                )
        else:
            log.debug3("Searching for all controllers...")
            for attr in dir(module):
                attr_val = getattr(module, attr)
                if cls._is_controller_type(attr_val):
                    log.debug2("Found Controller: %s", attr_val)
                    controller_types.append(attr_val)

        assert controller_types, f"No Controllers found in [{module_name}]"
        return controller_types

    @staticmethod
    def _parse_resource_dir(resource_dir: Optional[str]):
        """If given, this will parse all yaml files found in the given directory"""
        all_resources = []
        if resource_dir is not None:
            for fname in os.listdir(resource_dir):
                if fname.endswith(".yaml") or fname.endswith(".yml"):
                    resource_path = os.path.join(resource_dir, fname)
                    log.debug3("Reading resource file [%s]", resource_path)
                    with open(resource_path, encoding="utf-8") as handle:
                        all_resources.extend(yaml.safe_load_all(handle))
        return all_resources

    @staticmethod
    def _setup_watches(
        controller_types: List[Type[Controller]],
        resources: List[dict],
    ) -> Optional[DryRunDeployManager]:
        """Set up watches for all controllers. If in dry run mode, the
        DryRunDeployManager will be returned.
        """
        deploy_manager = None
        extra_kwargs = {}
        if config.dry_run:
            log.info("Running DRY RUN")
            deploy_manager = DryRunDeployManager(resources=resources)
            wm_type = watch_manager.DryRunWatchManager
            extra_kwargs["deploy_manager"] = deploy_manager
        elif config.watch_manager == "ansible":  # pragma: no cover
            log.info("Running Ansible Operator")
            wm_type = watch_manager.AnsibleWatchManager
        elif config.watch_manager == "python":  # pragma: no cover
            log.info("Running Python Operator")
            wm_type = watch_manager.PythonWatchManager
        else:
            raise ConfigError(f"Unknown watch manager {config.watch_manager}")

        for controller_type in controller_types:
            log.debug("Registering watch for %s", controller_type)
            wm_type(controller_type=controller_type, **extra_kwargs)
        return deploy_manager



"""
This DeployManager is responsible for delegating cluster operations to the
openshift library. It is the one that will be used when the operator is running
in the cluster or outside the cluster making live changes.
"""
# Standard
from collections import namedtuple
from typing import Callable, Iterator, List, Optional, Tuple
import copy
import threading
import time

# Third Party
from kubernetes import client
from kubernetes.watch import Watch
from openshift.dynamic import DynamicClient
from openshift.dynamic.apply import LAST_APPLIED_CONFIG_ANNOTATION, recursive_diff
from openshift.dynamic.exceptions import (
    ConflictError,
    ForbiddenError,
    NotFoundError,
    ResourceNotFoundError,
    ResourceNotUniqueError,
    UnprocessibleEntityError,
)
from openshift.dynamic.resource import Resource
import kubernetes
import urllib3

# First Party
from aconfig import Config
import alog

# Local
from .. import config
from .. import status as oper8_status
from ..exceptions import assert_cluster
from ..managed_object import ManagedObject
from ..verify_resources import verify_subsystem
from .base import DeployManagerBase, DeployMethod
from .kube_event import KubeEventType, KubeWatchEvent
from .owner_references import update_owner_references
from .replace_utils import requires_replace

log = alog.use_channel("OSFTD")

## Deploy Manager ##############################################################


# See this document for value reasonings
# https://github.com/kubernetes-client/python/blob/master/examples/watch/timeout-settings.md
SERVER_WATCH_TIMEOUT = 3600
CLIENT_WATCH_TIMEOUT = 30


class OpenshiftDeployManager(DeployManagerBase):
    """This DeployManager uses the openshift DynamicClient to interact with the
    cluster
    """

    def __init__(
        self,
        manage_ansible_status: bool = False,
        owner_cr: Optional[dict] = None,
    ):
        """
        Args:
            manage_ansible_status:  bool
                If true, oper8 will emulate the status management done natively
                by ansible based on the readiness values of oper8's native status
                management
            owner_cr:  Optional[dict]
                The dict content of the CR that triggered this reconciliation.
                If given, deployed objects will have an ownerReference added to
                assign ownership to this CR instance.
        """
        self.manage_ansible_status = manage_ansible_status
        self._owner_cr = owner_cr

        # Set up the client
        log.debug("Initializing openshift client")
        self._client = None

        # Keep a threading lock for performing status updates. This is necessary
        # to avoid running into 409 Conflict errors if concurrent threads are
        # trying to perform status updates
        self._status_lock = threading.Lock()

    @property
    def client(self):
        """Lazy property access to the client"""
        if self._client is None:
            self._client = self._setup_client()
        return self._client

    @alog.logged_function(log.debug)
    def deploy(
        self,
        resource_definitions: List[dict],
        manage_owner_references: bool = True,
        retry_operation: bool = True,
        method: DeployMethod = DeployMethod.DEFAULT,
        **_,  # Accept any kwargs to compatibility
    ) -> Tuple[bool, bool]:
        """Deploy using the openshift client

        Args:
            resource_definitions:  list(dict)
                List of resource object dicts to apply to the cluster
            manage_owner_references:  bool
                If true, ownerReferences for the parent CR will be applied to
                the deployed object

        Returns:
            success:  bool
                True if deploy succeeded, False otherwise
            changed:  bool
                Whether or not the deployment resulted in changes
        """
        return self._retried_operation(
            resource_definitions,
            self._apply,
            max_retries=config.deploy_retries if retry_operation else 0,
            manage_owner_references=manage_owner_references,
            method=method,
        )

    @alog.logged_function(log.debug)
    def disable(self, resource_definitions: List[dict]) -> Tuple[bool, bool]:
        """The disable process is the same as the deploy process, but the child
        module params are set to 'state: absent'

        Args:
            resource_definitions:  list(dict)
                List of resource object dicts to apply to the cluster

        Returns:
            success:  bool
                True if deploy succeeded, False otherwise
            changed:  bool
                Whether or not the delete resulted in changes
        """
        return self._retried_operation(
            resource_definitions,
            self._disable,
            max_retries=config.deploy_retries,
            manage_owner_references=False,
        )

    def get_object_current_state(
        self,
        kind: str,
        name: str,
        namespace: Optional[str] = None,
        api_version: Optional[str] = None,
    ) -> Tuple[bool, dict]:
        """The get_current_objects function fetches the current state using
        calls directly to the api client

        Args:
            kind:  str
                The kind of the object ot fetch
            name:  str
                The full name of the object to fetch
            namespace:  Optional[str]
                The namespace to search for the object or None for no namespace
            api_version:  Optional[str]
                The api_version of the resource kind to fetch

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  dict or None
                The dict representation of the current object's configuration,
                or None if not present
        """

        # Use the lazy discovery tool to first get all objects of the given type
        # in the given namespace, then look for the specific resource by name
        resources = self._get_resource_handle(kind, api_version)
        if not resources:
            return True, None

        if not namespace:
            resources.namespaced = False

        try:
            resource = resources.get(name=name, namespace=namespace)
        except ForbiddenError:
            log.debug(
                "Fetching objects of kind [%s] forbidden in namespace [%s]",
                kind,
                namespace,
            )
            return False, None
        except NotFoundError:
            log.debug(
                "No object named [%s/%s] found in namespace [%s]", kind, name, namespace
            )
            return True, None

        # If the resource was found, return it's dict representation
        return True, resource.to_dict()

    def watch_objects(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        api_version: Optional[str] = None,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
        resource_version: Optional[str] = None,
        watch_manager: Optional[Watch] = None,
    ) -> Iterator[KubeWatchEvent]:
        watch_manager = watch_manager if watch_manager else Watch()
        resource_handle = self._get_resource_handle(kind, api_version)
        assert_cluster(
            resource_handle,
            (
                "Failed to fetch resource handle for "
                + f"{namespace}/{api_version}/{kind}"
            ),
        )

        resource_version = resource_version if resource_version else 0

        while True:
            try:
                for event_obj in watch_manager.stream(
                    resource_handle.get,
                    resource_version=resource_version,
                    namespace=namespace,
                    name=name,
                    label_selector=label_selector,
                    field_selector=field_selector,
                    serialize=False,
                    timeout_seconds=SERVER_WATCH_TIMEOUT,
                    _request_timeout=CLIENT_WATCH_TIMEOUT,
                ):
                    event_type = KubeEventType(event_obj["type"])
                    event_resource = ManagedObject(event_obj["object"])
                    yield KubeWatchEvent(event_type, event_resource)
            except client.exceptions.ApiException as exception:
                if exception.status == 410:
                    log.debug2(
                        f"Resource age expired, restarting watch {kind}/{api_version}"
                    )
                    resource_version = None
                else:
                    log.info("Unknown ApiException received, re-raising")
                    raise exception
            except urllib3.exceptions.ReadTimeoutError:
                log.debug4(
                    f"Watch Socket closed, restarting watch {kind}/{api_version}"
                )
            except urllib3.exceptions.ProtocolError:
                log.debug2(
                    f"Invalid Chunk from server, restarting watch {kind}/{api_version}"
                )

            # This is hidden attribute so probably not best to check
            if watch_manager._stop:  # pylint: disable=protected-access
                log.debug(
                    "Internal watch stopped. Stopping deploy manager watch for %s/%s",
                    kind,
                    api_version,
                )
                return

    def filter_objects_current_state(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        namespace: Optional[str] = None,
        api_version: Optional[str] = None,
        label_selector: Optional[str] = None,
        field_selector: Optional[str] = None,
    ) -> Tuple[bool, List[dict]]:
        """The filter_objects_current_state function fetches a list of objects
        that match either/both the label or field selector
        Args:
            kind:  str
                The kind of the object to fetch
            namespace:  str
                The namespace to search for the object
            api_version:  str
                The api_version of the resource kind to fetch
            label_selector:  str
                The label_selector to filter the resources
            field_selector:  str
                The field_selector to filter the resources

        Returns:
            success:  bool
                Whether or not the state fetch operation succeeded
            current_state:  List[dict]
                A list of  dict representations for the objects configuration,
                or an empty list if no objects match
        """
        # Use the lazy discovery tool to first get all objects of the given type
        # in the given namespace, then look for the specific resource by name
        resources = self._get_resource_handle(kind, api_version)
        if not resources:
            return True, []

        if not namespace:
            resources.namespaced = False

        try:
            list_obj = resources.get(
                label_selector=label_selector,
                field_selector=field_selector,
                namespace=namespace,
            )
        except ForbiddenError:
            log.debug(
                "Fetching objects of kind [%s] forbidden in namespace [%s]",
                kind,
                namespace,
            )
            return False, []
        except NotFoundError:
            log.debug(
                "No objects of kind [%s] found in namespace [%s]", kind, namespace
            )
            return True, []

        # If the resource was found, get it's dict representation
        resource_list = list_obj.to_dict().get("items", [])
        return True, resource_list

    def set_status(  # pylint: disable=too-many-arguments
        self,
        kind: str,
        name: str,
        namespace: Optional[str],
        status: dict,
        api_version: Optional[str] = None,
    ) -> Tuple[bool, bool]:
        """Set the status in the cluster manifest for an object managed by this
        operator

        Args:
            kind:  str
                The kind of the object ot fetch
            name:  str
                The full name of the object to fetch
            namespace:  Optional[str]
                The namespace to search for the object.
            status:  dict
                The status object to set onto the given object
            api_version:  Optional[str]
                The api_version of the resource to update

        Returns:
            success:  bool
                Whether or not the status update operation succeeded
            changed:  bool
                Whether or not the status update resulted in a change
        """
        # Create a dummy resource to use in the common retry function
        resource_definitions = [
            {
                "kind": kind,
                "apiVersion": api_version,
                "metadata": {
                    "name": name,
                    "namespace": namespace,
                },
            }
        ]

        # Run it with retries
        return self._retried_operation(
            resource_definitions,
            self._set_status,
            max_retries=config.deploy_retries,
            status=status,
            manage_owner_references=False,
        )

    ## Implementation Helpers ##################################################

    @staticmethod
    def _setup_client():
        """Create a DynamicClient that will work based on where the operator is
        running
        """
        # Try in-cluster config
        try:
            log.debug2("Running with in-cluster config")

            # Create Empty Config and load in-cluster information
            kube_config = kubernetes.client.Configuration()
            kubernetes.config.load_incluster_config(client_configuration=kube_config)

            # Generate ApiClient and return Openshift DynamicClient
            api_client = kubernetes.client.ApiClient(kube_config)
            return DynamicClient(api_client)

        # Fall back to out-of-cluster config
        except kubernetes.config.ConfigException:
            log.debug2("Running with out-of-cluster config")
            return DynamicClient(kubernetes.config.new_client_from_config())

    @staticmethod
    def _strip_last_applied(resource_definitions):
        """Make sure that the last-applied annotation is not present in any of
        the resources. This can lead to recursive nesting!
        """
        for resource_definition in resource_definitions:
            last_applied = (
                resource_definition.get("metadata", {})
                .get("annotations", {})
                .get(LAST_APPLIED_CONFIG_ANNOTATION)
            )
            if last_applied:
                log.debug3("Removing [%s]", LAST_APPLIED_CONFIG_ANNOTATION)
                del resource_definition["metadata"]["annotations"][
                    LAST_APPLIED_CONFIG_ANNOTATION
                ]
                if not resource_definition["metadata"]["annotations"]:
                    del resource_definition["metadata"]["annotations"]

    def _get_resource_handle(self, kind: str, api_version: str) -> Optional[Resource]:
        """Get the openshift resource handle for a specified kind and api_version"""
        resources = None
        try:
            resources = self.client.resources.get(kind=kind, api_version=api_version)
        except (ResourceNotFoundError, ResourceNotUniqueError):
            try:
                resources = self.client.resources.get(
                    short_names=[kind], api_version=api_version
                )
            except (ResourceNotFoundError, ResourceNotUniqueError):
                log.debug(
                    "No objects of kind [%s] found or multiple objects matching request found",
                    kind,
                )
        return resources

    def _update_owner_references(self, resource_definitions):
        """If configured to do so, add owner references to the given resources"""
        if self._owner_cr:
            for resource_definition in resource_definitions:
                update_owner_references(self, self._owner_cr, resource_definition)

    def _retried_operation(
        self,
        resource_definitions,
        operation,
        max_retries,
        manage_owner_references,
        **kwargs,
    ):
        """Shared wrapper for executing a client operation with retries"""

        # Make sure the resource_definitions is a list
        assert isinstance(
            resource_definitions, list
        ), "Programming Error: resource_definitions is not a list"
        log.debug3("Running module with %d retries", max_retries)

        # If there are no resource definitions given, consider it a success with
        # no change
        if not resource_definitions:
            log.debug("Nothing to do for an empty list of resources")
            return True, False

        # Strip out last-applied annotations from all resources to avoid nested
        # annotations
        self._strip_last_applied(resource_definitions)

        # Add owner references if configured to do so
        if manage_owner_references:
            self._update_owner_references(resource_definitions)

        # Run each resource individually so that we can track partial completion
        success = True
        changed = False
        for resource_definition in resource_definitions:
            # Perform the operation and update the aggregate changed status
            try:
                changed = (
                    self._run_individual_operation_with_retries(
                        operation,
                        max_retries,
                        resource_definition=resource_definition,
                        **kwargs,
                    )
                    or changed
                )

            # On failure, mark it and stop processing the rest of the resources.
            # This is done because the resources in the file are assumed to be
            # in an intentional sequence and resources later in the file may
            # depend on resources earlier in the file.
            except Exception as err:  # pylint: disable=broad-except
                log.warning(
                    "Operation [%s] failed to execute: %s",
                    operation,
                    err,
                    exc_info=True,
                )
                success = False
                break

        # Return the aggregate success and change values
        return success, changed

    def _run_individual_operation_with_retries(
        self,
        operation: Callable,
        remaining_retries: int,
        resource_definition: dict,
        **kwargs,
    ):
        """Helper to execute a single helper operation with retries

        Args:
            operation:  Callable
                The operation function to run
            remaining_retries:  int
                The number of remaining retries
            resource_definition:  dict
                The dict representation of the resource being applied
            **kwargs:  dict
                Keyword args to pass to the operation beyond resource_definition

        Returns:
            changed:  bool
                Whether or not the operation resulted in meaningful change
        """
        try:
            return operation(resource_definition=resource_definition, **kwargs)
        except ConflictError as err:
            log.debug2("Handling ConflictError: %s", err)

            # If we have retries left, try again
            if remaining_retries:
                # Sleep for the backoff duration
                backoff_duration = config.retry_backoff_base_seconds * (
                    config.deploy_retries - remaining_retries + 1
                )
                log.debug3("Retrying in %fs", backoff_duration)
                time.sleep(backoff_duration)

                # Fetch the current resourceVersion and update in the
                # resource definition
                # NOTE: This can overwrite changes made external to the operator
                #   but that's an acceptable case since resources managed by
                #   oper8 should only be managed by oper8. In the rare case where
                #   oper8 shares ownership of a resource, any conflicts should
                #   be resoled cleanly on the next reconciliation.
                res_id = self._get_resource_identifiers(resource_definition)
                api_version = res_id.api_version
                kind = res_id.kind
                name = res_id.name
                namespace = res_id.namespace
                success, content = self.get_object_current_state(
                    kind=kind,
                    name=name,
                    namespace=namespace,
                    api_version=api_version,
                )
                assert_cluster(
                    success and content is not None,
                    (
                        "Failed to fetch updated resourceVersion for "
                        + f"{namespace}/{api_version}/{kind}/{name}"
                    ),
                )
                updated_resource_version = content.get("metadata", {}).get(
                    "resourceVersion"
                )
                assert_cluster(
                    updated_resource_version is not None,
                    "No updated resource version found!",
                )
                log.debug3(
                    "Updating resourceVersion from %s -> %s",
                    resource_definition.get("metadata", {}).get("resourceVersion"),
                    updated_resource_version,
                )
                resource_definition.setdefault("metadata", Config({}))[
                    "resourceVersion"
                ] = updated_resource_version

                # Run the retry
                log.debug3("Retrying")
                return self._run_individual_operation_with_retries(
                    operation, remaining_retries - 1, resource_definition, **kwargs
                )
            raise

    _ANSIBLE_COND_TYPE = "Running"
    _ANSIBLE_COND_RES_READY = {"ok": 1, "changed": 0, "skipped": 0, "failures": 0}
    _ANSIBLE_COND_RES_UNREADY = {"ok": 0, "changed": 0, "skipped": 0, "failures": 0}

    def _inject_ansible_status(self, status, previous_status):
        """If manage_ansible_status is enabled, this will inject the right
        ansible status values to emulate the format that ansible natively
        supports
        """
        previous_status = previous_status or {}

        # Check if the oper8 status indicates readiness
        is_ready = verify_subsystem(
            {"status": status}, desired_version=oper8_status.get_version(status)
        )
        prev_is_ready = verify_subsystem(
            {"status": previous_status},
            desired_version=oper8_status.get_version(previous_status),
        )
        log.debug3(
            "Status shows ready? %s. Previous ready? %s", is_ready, prev_is_ready
        )

        # Create the ansible status blob
        ansible_result = (
            self._ANSIBLE_COND_RES_READY if is_ready else self._ANSIBLE_COND_RES_UNREADY
        )
        log.debug3("Ansible Result: %s", ansible_result)

        # Determine if the condition has changed to know whether this is a
        # transition time
        current_ready_timestamp = oper8_status.get_condition(
            oper8_status.READY_CONDITION, status
        ).get(oper8_status.TIMESTAMP_KEY)
        prev_ready_timestamp = oper8_status.get_condition(
            oper8_status.READY_CONDITION, previous_status
        ).get(oper8_status.TIMESTAMP_KEY)
        if prev_ready_timestamp is not None and prev_is_ready == is_ready:
            log.debug3("No readiness change. Not a transition.")
            transition_time = prev_ready_timestamp
        else:
            log.debug3(
                "Transitioning from Ready(%s) -> Ready(%s)", prev_is_ready, is_ready
            )
            transition_time = current_ready_timestamp

        # Inject the final ansible condition
        conditions = [
            cond
            for cond in status.get("conditions", [])
            if cond.get("type") != self._ANSIBLE_COND_TYPE
        ]
        conditions.append(
            {
                "type": self._ANSIBLE_COND_TYPE,
                "ansibleResult": ansible_result,
                "lastTransitionTime": transition_time,
            }
        )
        status["conditions"] = conditions
        log.debug4("Status With Ansible: %s", status)

    @classmethod
    def _manifest_diff(cls, manifest_a, manifest_b) -> bool:
        """Helper to compare two manifests for meaningful diff while ignoring
        fields that always change.

        Returns:
            [bool, bool]: The first bool identifies if the resource changed while the
        """
        manifest_a = copy.deepcopy(manifest_a)
        manifest_b = copy.deepcopy(manifest_b)
        for metadata_field in [
            "resourceVersion",
            "generation",
            "managedFields",
            "uid",
            "creationTimestamp",
        ]:
            manifest_a.get("metadata", {}).pop(metadata_field, None)
            manifest_b.get("metadata", {}).pop(metadata_field, None)

        cls._strip_last_applied([manifest_b, manifest_a])
        diff = recursive_diff(
            manifest_a,
            manifest_b,
        )
        change = bool(diff)
        log.debug2("Found change? %s", change)
        log.debug3("A: %s", manifest_a)
        log.debug3("B: %s", manifest_b)
        return change

    # Internal struct to hold the key resource identifier elements
    _ResourceIdentifiers = namedtuple(
        "ResourceIdentifiers", ["api_version", "kind", "name", "namespace"]
    )

    @classmethod
    def _get_resource_identifiers(cls, resource_definition, require_api_version=True):
        """Helper for getting the required parts of a single resource definition"""
        api_version = resource_definition.get("apiVersion")
        kind = resource_definition.get("kind")
        name = resource_definition.get("metadata", {}).get("name")
        namespace = resource_definition.get("metadata", {}).get("namespace")
        assert None not in [
            kind,
            name,
        ], "Cannot apply resource without kind or name"
        assert (
            not require_api_version or api_version is not None
        ), "Cannot apply resource without apiVersion"
        return cls._ResourceIdentifiers(api_version, kind, name, namespace)

    ################
    ## Operations ##
    ################

    def _replace_resource(self, resource_definition: dict) -> dict:
        """Helper function to forcibly replace a resource on the cluster"""
        # Get the key elements of the resource
        res_id = self._get_resource_identifiers(resource_definition)
        api_version = res_id.api_version
        kind = res_id.kind
        name = res_id.name
        namespace = res_id.namespace

        # Strip out managedFields to let the sever set them
        resource_definition["metadata"]["managedFields"] = None

        # Get the resource handle
        log.debug2("Fetching resource handle [%s/%s]", api_version, kind)
        resource_handle = self._get_resource_handle(api_version=api_version, kind=kind)
        assert_cluster(
            resource_handle,
            (
                "Failed to fetch resource handle for "
                + f"{namespace}/{api_version}/{kind}"
            ),
        )

        log.debug2(
            "Attempting to put [%s/%s/%s] in %s",
            api_version,
            kind,
            name,
            namespace,
        )
        return resource_handle.replace(
            resource_definition,
            name=name,
            namespace=namespace,
            field_manager="oper8",
        ).to_dict()

    def _apply_resource(self, resource_definition: dict) -> dict:
        """Helper function to apply a single resource to the cluster"""
        # Get the key elements of the resource
        res_id = self._get_resource_identifiers(resource_definition)
        api_version = res_id.api_version
        kind = res_id.kind
        name = res_id.name
        namespace = res_id.namespace

        # Strip out managedFields to let the sever set them
        resource_definition["metadata"]["managedFields"] = None

        # Get the resource handle
        log.debug2("Fetching resource handle [%s/%s]", api_version, kind)
        resource_handle = self._get_resource_handle(api_version=api_version, kind=kind)
        assert_cluster(
            resource_handle,
            (
                "Failed to fetch resource handle for "
                + f"{namespace}/{api_version}/{kind}"
            ),
        )

        log.debug2(
            "Attempting to apply [%s/%s/%s] in %s",
            api_version,
            kind,
            name,
            namespace,
        )
        try:
            return resource_handle.server_side_apply(
                resource_definition,
                name=name,
                namespace=namespace,
                field_manager="oper8",
            ).to_dict()
        except ConflictError:
            log.debug(
                "Overriding field manager conflict for [%s/%s/%s] in %s ",
                api_version,
                kind,
                name,
                namespace,
            )
            return resource_handle.server_side_apply(
                resource_definition,
                name=name,
                namespace=namespace,
                field_manager="oper8",
                force_conflicts=True,
            ).to_dict()

    def _apply(self, resource_definition, method: DeployMethod):
        """Apply a single resource to the cluster

        Args:
            resource_definition:  dict
                The resource manifest to apply

        Returns:
            changed:  bool
                Whether or not the apply resulted in a meaningful change
        """
        changed = False

        res_id = self._get_resource_identifiers(resource_definition)
        api_version = res_id.api_version
        kind = res_id.kind
        name = res_id.name
        namespace = res_id.namespace

        # Get the current resource state
        success, current = self.get_object_current_state(
            kind=kind,
            name=name,
            namespace=namespace,
            api_version=api_version,
        )
        assert_cluster(
            success,
            (
                "Failed to fetch current state for "
                + f"{namespace}/{api_version}/{kind}/{name}"
            ),
        )
        if not current:
            current = {}

        # Determine if there will be a meaningful change (ignoring fields that
        # always change)
        changed = self._manifest_diff(current, resource_definition)

        # If there is meaningful change, apply this instance
        if changed:

            req_replace = requires_replace(current, resource_definition)

            # If the resource requires a replace operation then use put. Otherwise use
            # server side apply
            if (
                req_replace or method == DeployMethod.REPLACE
            ) and method != DeployMethod.UPDATE:
                apply_res = self._replace_resource(
                    resource_definition,
                )
            else:
                try:
                    apply_res = self._apply_resource(resource_definition)
                except UnprocessibleEntityError as err:
                    log.debug3("Caught 422 error: %s", err, exc_info=True)
                    if config.deploy_unprocessable_put_fallback:
                        log.debug("Falling back to PUT on 422: %s", err)
                        apply_res = self._replace_resource(
                            resource_definition,
                        )
                    else:
                        raise

            # Recompute the diff to determine if the apply actually caused a
            # meaningful change. This may have a different result than the check
            # above because the applied manifest does not always result in the
            # resource looking identical (e.g. removing field from applied =
            # manifest does not delete from the resource).
            changed = self._manifest_diff(current, apply_res)

        return changed

    def _disable(self, resource_definition):
        """Disable a single resource to the cluster if it exists

        Args:
            resource_definition:  dict
                The resource manifest to disable

        Returns:
            changed:  bool
                Whether or not the disable resulted in a meaningful change
        """
        changed = False

        # Get the key elements of the resource
        res_id = self._get_resource_identifiers(resource_definition)
        api_version = res_id.api_version
        kind = res_id.kind
        name = res_id.name
        namespace = res_id.namespace

        # Get the resource handle, handling missing kinds as success without
        # change
        log.debug2("Fetching resource [%s/%s]", api_version, kind)
        try:
            # Get a handle to the kind. This may fail with ResourceNotFoundError
            resource_handle = self.client.resources.get(
                api_version=api_version, kind=kind
            )

            # If resource is not namespaced set kubernetes api namespaced to false
            if not namespace:
                resource_handle.namespaced = False

            # Attempt to delete this instance. This ay fail with NotFoundError
            log.debug2(
                "Attempting to delete [%s/%s/%s] from %s",
                api_version,
                kind,
                name,
                namespace,
            )
            resource_handle.delete(name=name, namespace=namespace)
            changed = True

        # If the kind or instance is not found, that's a success without change
        except (ResourceNotFoundError, NotFoundError) as err:
            log.debug2("Valid error caught when disabling [%s/%s]: %s", kind, name, err)

        return changed

    def _set_status(self, resource_definition, status):
        """Disable a single resource to the cluster if it exists

        Args:
            resource_definition:  dict
                A dummy manifest holding the resource identifiers
            status:  dict
                The status to apply

        Returns:
            changed:  bool
                Whether or not the status update resulted in a meaningful change
        """
        changed = False

        # Get the key elements of the resource
        res_id = self._get_resource_identifiers(
            resource_definition, require_api_version=False
        )
        api_version = res_id.api_version
        kind = res_id.kind
        name = res_id.name
        namespace = res_id.namespace
        resource_handle = self.client.resources.get(api_version=api_version, kind=kind)

        # If resource is not namespaced set kubernetes api namespaced to false
        if not namespace:
            resource_handle.namespaced = False

        with self._status_lock:
            # Get the resource itself
            resource = resource_handle.get(name=name, namespace=namespace).to_dict()

            # Get the previous status and compare with the proposed status
            log.debug2(
                "Resource version: %s",
                resource.get("metadata", {}).get("resourceVersion"),
            )
            previous_status = resource.get("status")
            if previous_status == status:
                log.debug("Status has not changed. No update")

            else:
                # Inject the ansible status if enabled
                if self.manage_ansible_status:
                    log.debug2("Injecting ansible status")
                    self._inject_ansible_status(status, previous_status)

                # Overwrite the status
                resource["status"] = status
                resource_handle.status.replace(body=resource).to_dict()
                log.debug2(
                    "Successfully set the status for [%s/%s] in %s",
                    kind,
                    name,
                    namespace,
                )
                changed = True

            return changed



"""
Graph holds information about a Directed Acyclic Graph
"""


# Standard
from typing import Callable, List, Optional

# First Party
import alog

# Local
from .node import Node

log = alog.use_channel("DAG")

## Graph Class ##############################################################


class Graph:
    """Class for representing an instance of a Graph. Handles adding and removing nodes
    as well as graph functions like flattening"""

    def __init__(self) -> None:
        self.__node_dict = {}

        # Add the root node of the Graph. Every member of this graph is also a child
        # of the root node
        self.__root_node = Node()
        self.__node_dict[self.__root_node.get_name()] = self.__root_node

    ## Properties ##############################################################

    @property
    def root(self) -> Node:  # pylint: disable=invalid-name
        """The root node of the Graph"""
        return self.__root_node

    @property
    def node_dict(self) -> dict:
        """Dictionary of all node names and their nodes"""
        return self.__node_dict

    ## Modifiers ##############################################################

    def add_node(self, node: Node):
        """Add node to graph
        Args:
            node:  Node
                The node to be added to the Dag.
        """
        if not node.get_name():
            raise ValueError("None is reserved for the root node of the dag Graph")

        if node.get_name() in self.node_dict:
            raise ValueError(
                f"Only one node with id {node.get_name()} can be added to a Graph"
            )

        self.node_dict[node.get_name()] = node
        self.root.add_child(node)

    def add_node_dependency(
        self, parent_node: Node, child_node: Node, edge_fn: Optional[Callable] = None
    ):
        """Add dependency or "edge" to graph between two nodes. This is the same
        as doing parent_node.add_dependency(child_node)
        Args:
            parent_node:  Node
                The parent or dependent node aka the node that must wait
            child_node: Node
                The child or dependency node aka the node that must be deployed first
            edge_fn:
        """
        if not self.get_node(parent_node.get_name()):
            raise ValueError(f"Parent node {parent_node} is not present in Graph")

        if not self.get_node(child_node.get_name()):
            raise ValueError(f"Child node {child_node} is not present in Graph")

        # Make sure edits are applied to the nodes already present in the graph
        parent_node = self.get_node(parent_node.get_name())
        child_node = self.get_node(child_node.get_name())

        parent_node.add_child(child_node, edge_fn)

    ## Accessors ##############################################################

    def get_node(self, name: str):  # pylint: disable=invalid-name
        """Get the node with name"""
        return self.node_dict.get(name)

    def get_all_nodes(self):
        """Get list of all nodes"""
        return [node for node, _ in self.root.get_children()]

    def has_node(self, node: Node):  # pylint: disable=invalid-name
        """Check if node is in graph"""
        return self.root.has_child(node)

    def empty(self):
        """Check if a graph is empty"""
        return len(self.root.get_children()) == 0

    ## Graph Functions ##############################################################

    def topology(self) -> List["Node"]:
        """Get a list of nodes in deployment order"""
        topology = self.root.topology()
        topology.remove(self.root)
        return topology

    ## Internal Functions ##############################################################

    def __repr__(self):
        str_list = []
        for child, _ in self.root.get_children():
            child_str_list = [node.get_name() for node, _ in child.get_children()]
            str_list.append(f"{child.get_name()}:[{','.join(child_str_list)}]")

        return f"Graph({{{','.join(str_list)}}})"

    def __contains__(self, item: Node):
        return self.has_node(item)

    def __iter__(self):
        """Iterate over all child nodes"""
        return self.get_all_nodes().__iter__()



"""
Helper module to define shared types related to Kube Events
"""

# Standard
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum

# First Party
import alog

# Local
from ..managed_object import ManagedObject

log = alog.use_channel("KUBEWATCH")


class KubeEventType(Enum):
    """Enum for all possible kubernetes event types"""

    DELETED = "DELETED"
    MODIFIED = "MODIFIED"
    ADDED = "ADDED"


@dataclass
class KubeWatchEvent:
    """DataClass containing the type, resource, and timestamp of a
    particular event"""

    type: KubeEventType
    resource: ManagedObject
    timestamp: datetime = field(default_factory=datetime.now)



# Standard
from unittest.mock import Mock

# Third Party
import pytest

# Local
from oper8 import Component, Controller, component, controller
from oper8.test_helpers.helpers import library_config, setup_cr, setup_session

## @component ##################################################################

mock_verify = Mock()


@component(name="pure_comp")
class PureComponent(Component):
    def __init__(self, session):
        super().__init__(session=session)
        self.add_resource(
            "test_obj",
            {
                "kind": "test",
                "apiVersion": "v1",
                "metadata": {"name": session.get_scoped_name("test_obj")},
            },
        )

        # Attach the verify mock
        self.verify = mock_verify.method


def test_component_creation():
    """Ensure that decorated components are created correctly"""

    session = setup_session()

    # Pure components are a Component
    comp = PureComponent(session)
    assert isinstance(comp, Component)

    # Pure components can be built
    config = comp.to_config(session)
    assert config[0].metadata.name == session.get_scoped_name("test_obj")

    # Pure components have a class attribute "name"
    assert hasattr(comp.__class__, "name")


## @controller #################################################################

setup_components_mock = Mock()


@controller(
    group="foo.bar",
    version="v1",
    kind="Foo",
)
class FooController(Controller):
    def setup_components(self, *_, **__):
        setup_components_mock()


def test_controller_creation():
    """Make sure that an @controller decorated class can be created"""

    # Set up the controller
    with library_config(dry_run=True):
        ctrlr = FooController()
    assert isinstance(ctrlr, Controller)

    # Check class properties
    assert ctrlr.__class__.group == "foo.bar"
    assert ctrlr.__class__.version == "v1"
    assert ctrlr.__class__.kind == "Foo"

    # Roll out
    ctrlr.run_reconcile(setup_session())


def test_controller_extra_properties():
    """Make sure that the extra_properties argument to @controller properly
    attaches the given properties to the Controller class
    """

    @controller(
        group="foo.bar",
        version="v2",
        kind="Bar",
        extra_properties={"foo": "bar"},
    )
    class BarController(Controller):
        def setup_components(self, *_, **__):
            setup_components_mock()

    assert hasattr(BarController, "foo")
    assert BarController.foo == "bar"


def test_controller_finalizer():
    """Make sure that the extra_properties argument to @controller properly
    attaches the given properties to the Controller class
    """

    @controller(
        group="foo.bar",
        version="v2",
        kind="Bar",
        finalizer="test",
    )
    class BarController(Controller):
        def setup_components(self, *_, **__):
            setup_components_mock()

    assert BarController.finalizer == "test"



"""This class provides a base class with shared functionality that all
concrete components can use.
"""

# Standard
from typing import Union

# First Party
import alog

# Local
from .utils import common, deps_annotation
from oper8 import Component, Session

log = alog.use_channel("OPCMP")


## Oper8xComponent #############################################################

default = {}


class Oper8xComponent(Component):
    """The Oper8xComponent provides common config-based utilities on top of the
    core oper8.Component base class. It can be used as a drop-in replacement.
    """

    def __init__(self, session: Session, disabled: bool = False):
        """Construct with a member to access the session in implementations

        Args:
            session:  Session
                The session for the current deployment
            disabled:  bool
                Whether or not this component is disabled in the current
                configuration
        """
        super().__init__(session=session, disabled=disabled)
        self._session = session

    @property
    def session(self):
        return self._session

    ## Interface Overrides #####################################################

    def deploy(self, session: Session) -> bool:
        """Override the base Component's implementation of deploy to insert the
        dependency hash annotation. See NOTE in deps_annotation for explanation
        of why deploy is used instead of update_object_definition.

        Args:
            session:  Session
                The session for the current deployment

        Returns:
            success:  bool
                True on successful application of the resource to the cluster
        """
        for obj in self.managed_objects:
            obj.definition = deps_annotation.add_deps_annotation(
                self, session, obj.definition
            )
        return super().deploy(session)

    def update_object_definition(
        self,
        session: Session,
        internal_name: str,
        resource_definition: dict,
    ) -> dict:
        """For components assigned to different namespaces, ensure that the
        target namespace is set

        Args:
            session:  Session
                The session for this deploy
            internal_name:  str
                The internal name of the object to update
            resource_definition:  dict
                The dict representation of the resource to modify

        Returns:
            resource_definition:  dict
                The dict representation of the resource with any modifications
                applied
        """

        # Call the base implementation
        resource_definition = super().update_object_definition(
            session,
            internal_name,
            resource_definition,
        )

        # Inject namespace override for this component if given
        namespace_override = session.config.get(self.name, {}).get("namespace")
        if namespace_override is not None:
            log.debug2("Namespace  override for %s: %s", self, namespace_override)
            metadata = resource_definition.get("metadata")
            assert isinstance(metadata, dict), "Resource metadata is not a dict!"
            metadata["namespace"] = namespace_override

        return resource_definition

    ## Shared Utilities ########################################################

    def get_cluster_name(self, resource_name: str) -> str:
        """Get the name for a given resource with any instance scoping applied

        Args:
            resource_name:  str
                The unscoped name of a kubernetes resource

        Returns:
            resource_cluster_name:  str
                The name that the resource will use in the cluster
        """
        return common.get_resource_cluster_name(
            resource_name=resource_name,
            component=self.name,
            session=self.session,
        )

    def get_replicas(self, force: bool = False) -> Union[int, None]:
        """Get the replica count for this component based on the current
        deploy's t-shirt size and the state of the instance-size label. A
        replica count is only returned if there is not an existing replica count
        in the cluster for this deployment, the t-shirt size has changed, or
        the force flag is True.

        Args:
            force: bool
                If True, the state of the cluster will not be checked

        Returns:
            replicas:  Union[int, None]
                If replicas should be set for this deployment, the integer count
                will be returned, otherwise None is returned.
        """
        return common.get_replicas(
            session=self.session,
            component_name=self.name,
            unscoped_name=self.name,
            force=force,
        )



"""Useful Constants"""


## Reconcile Constants

# Default timeout when joining processes
JOIN_PROCESS_TIMEOUT = 5

# Default Poll Time for running reconcile cleanup on shutdown
SHUTDOWN_RECONCILE_POLL_TIME = 0.1

## Timer Constants

# Minimum wait time between checks in periodic thread
MIN_SLEEP_TIME = 1


## Filter Constants

# Only keep a set number of resource versions per watched resource
# this limits the amount of memory used
RESOURCE_VERSION_KEEP_COUNT = 20

# List of reserved annotations used by the platforms
RESERVED_PLATFORM_ANNOTATIONS = [
    "k8s.io",
    "kubernetes.io",
    "openshift.io",
]
